"""Bind constraint"""

from __future__ import annotations

import logging
from functools import cached_property
from typing import TYPE_CHECKING

from ...utils.decorators import timer
# from ...components.temporal.modes import Modes
from ...utils.math import normalize

logger = logging.getLogger("energia")

if TYPE_CHECKING:
    from gana import P, V
    from gana.sets.constraint import C
    from gana.sets.function import F

    from ..._core._component import _Component
    from ..._core._x import _X
    from ..indices.sample import Sample


class Bind:
    """Bind constraint

    :param sample: The sample variable to bind
    :type sample: Sample
    :param parameter: The parameter bound
    :type parameter: float | list[float] | dict[float, float] | tuple[float, float] | list[tuple[float, float]]
    :param leq: If True, the sample is constrained to be less than or equal to the bound
    :type leq: bool
    :param geq: If True, the sample is constrained to be greater than or equal to the bound
    :type geq: bool
    :param eq: If True, the sample is constrained to be equal to the bound
    :type eq: bool
    :param forall: If provided, the constraint is applied for all elements in this list
    :type forall: list[_X | _Component] | None

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar nominal: The nominal value of the sample variable.
    :vartype nominal: float | None
    :ivar norm: If True, the sample variable is normalized.
    :vartype norm: bool
    :ivar domain: The domain of the sample.
    :vartype domain: Domain
    :ivar aspect: The aspect of the sample.
    :vartype aspect: Aspect
    :ivar report: If True, the sample variable is reported.
    :vartype report: bool
    :ivar program: The program to which the sample belongs.
    :vartype program: Prg
    """

    def __init__(
        self,
        sample: Sample,
        parameter: (
            float
            | list[float]
            | dict[float, float]
            | tuple[float, float]
            | list[tuple[float, float]]
        ),
        leq: bool = False,
        geq: bool = False,
        eq: bool = False,
        forall: list[_X | _Component] | None = None,
        parameter_name: str = "",
    ):
        self.sample = sample
        self._parameter, self.parameter_name = parameter, parameter_name
        self.leq, self.geq, self.eq = leq, geq, eq
        self.forall = forall

        self._handshake()

        if self.forall:
            # if as set is passed
            # write the constraint 'for all' elements in it
            self._write_forall()
            return

        if isinstance(self._parameter, dict):
            # if a dict is passed, it is assumed to be mode bounds
            if self.iscalc:
                self._calc_w_modes()
            else:
                self._write_w_modes()
            return

        self.write()

    @timer(logger, kind="bind")
    def write(self):
        """Writes the bind constraint"""

        # the lhs comes from the sample
        # calling the lhs here, updates the
        _ = self.lhs

        if self._check_existing():
            return False

        if self.leq:
            self.cons: C = self.lhs <= self.rhs

        elif self.eq:
            self.cons: C = self.lhs == self.rhs

        elif self.geq:
            self.cons: C = self.lhs >= self.rhs

        else:
            return False

        self._inform()

        # set the constraint
        setattr(
            self.program,
            self.cons_name,
            self.cons,
        )
        # returned for @timer
        return self.sample, self.rel

    @cached_property
    def parameter(self):
        """Parameter bound of the bind constraint"""

        if self.nominal:
            # if a nominal value for the self.parameter is passed
            # this is essentially the expectation
            # skipping an instance check here
            # if a non iterable is passed, let an error be raised
            if self.norm:
                _parameter = normalize(self._parameter)

            else:
                _parameter = self._parameter

            # if the sample needs to be normalized
            _parameter = [
                (
                    (self.nominal * i[0], self.nominal * i[1])
                    if isinstance(i, tuple)
                    else self.nominal * i
                )
                for i in _parameter
            ]
            return _parameter
        return self._parameter

    @cached_property
    def lhs(self):
        """Left hand side of the bind constraint"""

        # ------Get LHS
        # lhs needs to be determined here
        # because V will be spaced and timed if not passed by user
        # .X(), .Vb() need time and space
        return self.sample.V(self.parameter)

    @property
    def rhs(self) -> V | F | P:
        """Right hand side of the bind constraint"""
        if self.of:
            # if the dependent variable is not set, creates issues.
            # ------if a calculation is being done
            return self.parameter * self.of(*self.domain.index_spatiotemporal).V(
                self.parameter
            )

        if self.aspect.bound:
            # ------if variable bound
            if self.report:
                # ------if variable bound and reported
                # we do not want a bi-linear term
                return self.parameter * self.sample.X(self.parameter)

            # ------if just variable bound

            return self.parameter * self.sample.Vb()

        if self.report or self.domain.modes is not None:
            # ------if  self.parameter bound and reported or has modes
            # create reporting variable write v <= p*x
            self.aspect.update(self.domain, reporting=True)
            return self.parameter * self.sample.X(self.parameter)

        # ------if just self.parameter bound
        return self.parameter

    @cached_property
    def rel(self):
        """Constraint name suffix"""
        if self.leq:
            return "ub"
        elif self.eq:
            if self.of is not None:
                if self.report:
                    return "inc_calc"
                return "calc"
            return "eq"
        elif self.geq:
            return "lb"

    @cached_property
    def cons_name(self):
        """Constraint name"""
        return rf"{self.aspect.name}{self.domain.idxname}_{self.rel}"

    @property
    def iscalc(self) -> bool:
        """Is this a calculation bind constraint?"""
        return self.of is not None

    def _write_forall(self):
        """Writes the bind constraint for all elements in the set"""

        for n, idx in enumerate(self.forall):

            lhs = self.sample(idx)

            try:
                # if a list is passed
                # or any iterable vector
                rhs = self.parameter[n]
            except TypeError:
                # if not repeat the same value
                # over all elements
                rhs = self.parameter

            if self.leq:
                _ = lhs <= rhs
            if self.geq:
                _ = lhs >= rhs
            if self.eq:
                _ = lhs == rhs

    def _calc_w_modes(self):
        """Write with modes"""
        # if this is a dict, piece wise linear functions are being passed
        # Take the example of expenditure and capacity being modeled
        # sat the input is {(0, 200): 5000, (200, 300): 4000, (300, 400): 3000
        # what this implies is that for the capacity between 0 and 200, the expenditure is 5000
        # for the capacity between 200 and 300, the expenditure is 4000
        # for the capacity between 300 and 400, the expenditure is 3000
        # Modes objects index the bin. Three in this case : 1 - (0,200), 2 - (200,300), 3 - (300,400)
        # We need the following equations:
        # *1. capacity = capacity(bin0) + capacity(bin1) + capacity(bin2)
        # *2. x_capacity = 0*capacity(bin0) + 200*capacity(bin1) + 300*capacity(bin2), where is a reporting binary
        # *3-5. spend(bin0) = 5000*capacity(bin0); spend(bin1) = 4000*capacity(bin1); spend(bin2) = 3000*capacity(bin2)
        # *6. spend = spend(bin0) + spend(bin1) + spend(bin2)
        # this takes care of *1 and *2
        _ = self.of == dict(enumerate(self.parameter))
        # this takes care of *3-*6

        # the new modes object would have just been added to the model
        modes = self.model.modes[-1]

        _ = self.of(modes)[self.sample(modes)] == list(self.parameter.values())

    def _write_w_modes(self):
        """Writes the bind constraint with modes"""
        # create modes
        self.modes = self.model.Modes(size=len(self.parameter), sample=self.sample)

        mode_bounds = [
            (
                (self.parameter[i - 1], self.parameter[i])
                if i - 1 in self.parameter
                else (0, self.parameter[i])
            )
            for i in self.parameter
        ]

        _ = self.sample(self.modes) >= [b[0] for b in mode_bounds]
        _ = self.sample(self.modes) <= [b[1] for b in mode_bounds]

    def _check_existing(self) -> bool:
        """Checks if aspect already has been bound in that space"""
        if not self.iscalc:
            if (
                (self.domain.space, self.domain.time)
                in self.aspect.bound_spaces[self.domain.primary][self.rel]
            ) and not self.domain.modes:
                return True

            self.aspect.bound_spaces[self.domain.primary][self.rel].append(
                (self.domain.space, self.domain.time)
            )
        return False

    def _inform(self):
        """Informs the aspect and domain about the bind constraint"""

        # categorize the constraint
        if self.iscalc:
            self.cons.categorize("Calculations")
        elif self.domain.modes:
            self.cons.categorize("Piecewise Linear")
        else:
            self.cons.categorize("Binds")

        # let the aspect know about the new constraint
        if self.cons_name not in self.aspect.constraints:
            self.aspect.constraints.append(self.cons_name)

        # let all objects in the domain know that
        # a constraint with this name contains it
        self.domain.inform_indices(self.cons_name)

        self.model.scenario.update(self.sample, self.rel, self.parameter)

    def _handshake(self):
        """Borrow attributes from sample"""
        # borrowed from Sample
        self.model = self.sample.model
        self.nominal = self.sample.nominal
        self.norm = self.sample.norm
        self.domain = self.sample.domain
        self.aspect = self.sample.aspect
        self.report = self.sample.report
        self.program = self.sample.program
        self.of = self.sample.of



"""Edge"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..._core._name import _Name

if TYPE_CHECKING:
    from ...represent.ations.graph import Graph


class Edge(_Name):
    """
    Edge of a Graph

    :param label: Label of the component, used for plotting. Defaults to None.
    :type label: str, optional
    :param graph: Graph to which the edge belongs. Defaults to None.
    :type graph: Graph, optional
    :ivar name: name. Defaults to ''.
    :vartype name: str

    .. note::
        - name and Graph are set when made a Graph attribute.
    """

    def __init__(self, label: str = ""):
        self.graph: Graph | None = None
        _Name.__init__(self, label=label)



"""Utilities to perform mathematical operations"""

from math import erf, exp, pi, sqrt

import numpy


def norm_constant(p, mu, sigma) -> float:
    """
    Calculates the normal constant

    :param p: level of compliance
    :type p: float
    :param mu: mean
    :type mu: float
    :param sigma: standard deviation
    :type sigma: float
    :return: normal constant
    :rtype: float
    """
    x = mu + erf(1 / sqrt(2) * p) * sigma * sqrt(2)
    return 1 / (sigma * sqrt(2 * pi)) * exp(-((x - mu) ** 2) / (2 * sigma**2))


def find_euclidean_distance(a: list, b: list) -> float:
    """
    finds euclidean distances between two cluster nodes

    :param a: coordinates of cluster node a
    :type a: list
    :param b: coordinates of cluster node b
    :type b: list

    :returns: euclidean distance
    :rtype: float
    """
    distance_ = [(a - b) ** 2 for a, b in zip(a, b)]
    distance_ = sum(distance_)
    return distance_


def generate_connectivity_matrix(scale_len) -> numpy.array:
    """
    Generates a connectivity matrix to maintain chronology [..1,0,1..]

    :param scale_len: length of the scale
    :type scale_len: int

    :returns: connectivity matrix
    :rtype: numpy.array
    """
    connect_ = numpy.zeros((scale_len, scale_len), dtype=int)
    for i_ in range(len(connect_)):

        if i_ == 0:
            # connect_[i,364] = 1 #uncomment these to generate a cyclic matrix
            connect_[i_, 1] = 1
        elif i_ == scale_len - 1:
            connect_[i_, scale_len - 2] = 1
            # connect_[i,0] = 1
        else:
            connect_[i_, i_ - 1] = 1
            connect_[i_, i_ + 1] = 1
    return connect_


def normalize(data: list, how: str = "max") -> list:
    """
    min max normalization of data

    :param data: time-series data
    :type data: numpy.array | DataFrame
    :param how: min-max or max, defaults to "max"
    :type how: str, optional

    :return: normalized data
    :rtype: numpy.array | DataFrame
    """
    if all(isinstance(i, tuple) for i in data):
        # lower bound
        lb = [i[0] for i in data]
        # upper bound
        ub = [i[1] for i in data]
        # normalize lower and upper bounds individually
        lb_norm = normalize(lb, how=how)
        ub_norm = normalize(ub, how=how)
        return list(zip(lb_norm, ub_norm))

    if how == "min_max":
        min_ = min(data)
        max_ = max(data)
        return [(i - min_) / (max_ - min_) for i in data]

    if how == "max":
        return [i / max(data) for i in data]



"""Scenario"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..._core._hash import _Hash
from ...utils.dictionary import merge_trees

if TYPE_CHECKING:
    from ...modeling.indices.sample import Sample
    from ...represent.model import Model


class Scenario(_Hash):
    """Scenario representation"""

    def __init__(self, model: Model):

        self.model = model
        self.name = rf"Scenario({self.model})"

        # Bounds, upper, lower, and equality
        self.ubs = {}
        self.lbs = {}
        self.eqs = {}

        # Calculations
        self.calcs = {}
        # incidental calculations
        self.inc_calcs = {}

    @property
    def _(self):
        """Returns the scenario representation as a dictionary"""
        return {
            "ubs": self.ubs,
            "lbs": self.lbs,
            "eqs": self.eqs,
            "calcs": self.calcs,
            "inc_calcs": self.inc_calcs,
        }

    def update(
        self,
        sample: Sample,
        # domain: Domain,
        rel: str,
        parameter: float | list[float],
    ):
        """Update the scenario representation"""

        if rel == 'ub':
            self.ubs = merge_trees(
                self.ubs, {sample.aspect: sample.domain.param_tree(parameter)}
            )

        elif rel == 'lb':
            self.lbs = merge_trees(
                self.lbs, {sample.aspect: sample.domain.param_tree(parameter)}
            )
        elif rel == 'eq':
            self.eqs = merge_trees(
                self.eqs, {sample.aspect: sample.domain.param_tree(parameter)}
            )

        elif rel == 'calc':
            self.calcs = merge_trees(
                self.calcs, {sample.aspect: sample.domain.param_tree(parameter)}
            )
        elif rel == 'inc_calc':
            self.inc_calcs = merge_trees(
                self.inc_calcs, {sample.aspect: sample.domain.param_tree(parameter)}
            )



"""Recipes to create Aspects"""

from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Type

from .aspect import Aspect

if TYPE_CHECKING:
    from ..._core._component import _Component


@dataclass
class Recipe:
    """
    Recipe to create Aspects

    :param name: Name of the aspect
    :type name: str
    :param kind: Kind of aspect, e.g., Aspect, Control, etc.
    :type kind: Type[Aspect]
    :param primary_type: Primary type(s) of component the aspect can be applied to
    :type primary_type: tuple[Type[_Component]] | Type[_Component]
    :param label: Label for plotting. Defaults to "".
    :type label: str
    :param add: Add string for constraints. Defaults to "".
    :type add: str
    :param sub: Subtract string for constraints. Defaults to "".
    :type sub: str
    :param bound: Bound string for constraints. Defaults to "".
    :type bound: str
    :param ctrl: Control string for constraints. Defaults to "".
    :type ctrl: str
    :param ispos: If the aspect is positive. Defaults to True.
    :type ispos: bool
    :param nn: If the aspect is non-negative. Defaults to True.
    :type nn: bool
    :param latex: LaTeX representation. Defaults to "".
    :type latex: str

    :ivar args: Arguments for aspect creation.
    :vartype args: dict[str, Any]
    """

    name: str
    kind: Type[Aspect]
    primary_type: tuple[Type[_Component]] | Type[_Component]
    label: str = ""
    add: str = ""
    sub: str = ""
    bound: str = ""
    ctrl: str = ""
    ispos: bool = True
    nn: bool = True
    latex: str = ""

    def __post_init__(self):
        self.args = {
            k: v
            for k, v in asdict(self).items()
            if k not in ("name", "kind") and (isinstance(v, bool) or v)
        }

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



"""Sample of an Aspect"""

from __future__ import annotations

import logging
from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Self

from gana import I as Idx
from gana import V, inf, sigma, sup

from ...utils.dictionary import merge_trees
from ..constraints.bind import Bind

logger = logging.getLogger("energia")


if TYPE_CHECKING:
    from gana import P, Prg
    from gana.sets.constraint import C
    from gana.sets.function import F

    from ..._core._component import _Component
    from ..._core._x import _X
    from ..variables.aspect import Aspect
    from .domain import Domain


# ------------------------------------------------------------------------------
# Sample is the primary constraint generator
# These are essentially a variable waiting to be bound
# ------------------------------------------------------------------------------


class Sample:
    """
    Sets a bound on a variable (V) within a particular domain.

    :param aspect: Aspect to which the constraint is applied.
    :type aspect: Aspect | None
    :param domain: Domain over which the aspect is defined.
    :type domain: Domain | None
    :param timed: If the temporal index is predetermined. Defaults to None.
    :type timed: bool | None
    :param spaced: If the spatial index is predetermined. Defaults to None.
    :type spaced: bool | None

    :ivar name: Name of the bind.
    :vartype name: str | None
    :ivar model: Model to which the generator belongs.
    :vartype model: Model | None
    :ivar program: Gana Program to which the generated constraint belongs.
    :vartype program: Prg | None
    :ivar opr: Operation to bind (in lieu of a single variable). Defaults to None.
    :vartype opr: F | None
    :ivar domains: Set of domains over which the Bind is applied. Defaults to [].
    :vartype domains: Domain
    :ivar hasinc: If the Bind has some incidental calculation. Defaults to False.
    :vartype hasinc: bool

    .. note::
        - ``timed`` and ``spaced`` help skip the calculation of finding the appropriate index.
            For time, this is done based on the length of the input parameter.
        - ``opr`` is useful if providing a combined bound to different variables.
        - ``name`` is generated based on the variable.
        - ``domains`` are updated as the program is built.
    """

    def __init__(
        self,
        aspect: Aspect,
        domain: Domain,
        label: str = "",
        timed: bool = False,
        spaced: bool = False,
        report: bool = False,
    ):

        # this is the aspect for which the constraint is being defined
        self.aspect = aspect
        # the domain is passed when the aspect is called using __call__()
        self.domain = domain

        self.label = label
        # if the temporal index is predetermined
        self.timed = timed
        # if the spatial index is predetermined
        self.spaced = spaced

        self.model = self.aspect.model
        self.program = self.model.program
        self.balances = self.model.balances

        # if the aspect is bound (operate for example)
        self.bound = self.aspect.bound

        # this is set if the aspect needs a reporting binary variable
        self.report = report

        # if incidental calculation is generated
        self.hasinc: bool = False

        # if nominal is provided
        # and multiplied by the nominal value
        self.nominal: float | None = None
        # the input argument is normalized if True
        self.norm: bool = False

        # the bound is set for all indices
        self._forall: list[_X] = []

        # parameters
        self.parameter: P = None
        self.length: int = 0

    @property
    def of(self) -> Self | None:
        """Sample being calculated"""
        if self.domain.samples:
            return self.domain.samples[0]
        return None

    @property
    def name(self) -> str:
        """Name of the constraint"""
        return f"{self.domain.primary}.{self.aspect.name}"

    @property
    def index(self) -> list[_Component]:
        """_Index"""
        return self.domain.index

    @property
    def index_short(self) -> list[_X | Sample]:
        """Short Index"""
        return self.domain.index_short

    @cached_property
    def I(self) -> Idx:
        """gana index set (I)"""
        return self.domain.I

    @property
    def x(self) -> Self:
        """Binary Reporting Variable"""
        self.report = True
        return self

    @property
    def add(self) -> Self:
        """Positive Control Response"""
        return self.aspect.add(*self.index)

    @property
    def sub(self) -> Self:
        """Negative Control Response"""
        return self.aspect.sub(*self.index)

    @property
    def F(self):
        """Function"""
        if self.report:
            return self.X(1)
        return self.V(1)

    @property
    def constraints(self):
        """Constraints"""
        return self.aspect.constraints

    def show(self, descriptive=False):
        """Pretty print the component"""
        constraints = list(chain.from_iterable(i.constraints for i in self.index))

        for c in constraints:
            if c in self.aspect.constraints:
                cons: C = getattr(self.program, c)
                cons.show(descriptive)

    @cached_property
    def time(self):
        """Matches an appropriate temporal scale"""
        if not self.timed:

            if isinstance(self.parameter, list):
                # if list is given, find using length of the list
                if self.domain.modes is not None:
                    self.domain.periods = self.aspect.time.find(
                        len(self.parameter) / len(self.domain.modes),
                    )
                else:
                    self.domain.periods = self.aspect.time.find(len(self.parameter))

            elif isinstance(self.length, int):
                # if length is given, use it directly
                self.domain.periods = self.aspect.time.find(self.length)
            else:
                # else the size of parameter set is exactly one
                # or nothing is given, meaning the variable is not time dependent
                # thus, index by horizon
                self.domain.periods = self.aspect.horizon

        return self.domain.periods

    @cached_property
    def space(self):
        """Assigns to network is spatial index is not given"""
        if not self.spaced:
            # if spatial index is not explicity given
            # default to the network
            self.domain.location = self.aspect.network

        return self.domain.location

    def Vlag(self):
        """Handles lagged domains"""
        # with lag it is assumed that the variable of which this is a lagged subset is
        # already defined
        # for example, if opr_t = opr_t-1 + x_t, then opr_t is already defined
        try:
            return getattr(self.program, self.aspect.name)(*self.domain.I)

        except AttributeError:
            _ = self == True
            return getattr(self.program, self.aspect.name)(*self.domain.I)

        except KeyError:
            # the variable has not been defined yet
            lag = self.domain.lag
            self.domain = self.domain.change(
                {"lag": None, "periods": self.domain.lag.of},
            )
            args = (self.parameter, self.length)

            if self.hasinc:
                _ = self.Vinc(*args)

            if self.report:
                _ = self.X(*args)

            else:
                _ = self.V(*args)

            self.domain = self.domain.change({"lag": lag, "periods": None})
            return getattr(self.program, self.aspect.name)(*self.domain.I)

    def _inform(self):
        """Informs the aspect and domain about the sample"""
        # updated the indices
        # we can be confident that the self.I is unique
        # because of the check above
        self.aspect.indices.append(self.I)

        # this updates the balanced dictionary, by adding the commodity as a key

        if self.domain.commodity and not self.domain.lag:
            _ = self.balances[self.domain.commodity][self.domain.space][
                self.domain.time
            ]

        # this lets all self.I elements in the domain know
        # that the aspect was sampled
        self.domain.update_domains(self.aspect)

        # ------Update the disposition ---------------

        # get the primary component
        # update the disposition dictionary
        self.model.dispositions = merge_trees(
            self.model.dispositions,
            {self.aspect: self.domain.tree},
        )

        # for the same aspect, map variables with higher order indices
        # to variables with lower order indices
        self.aspect.update(self.domain)

        self.aspect.domains.append(self.domain)

    def _init_V(self, parameter: float | list = None, length: int = None):
        """Initialize making a variable"""

        self.parameter = parameter
        self.length = length

        if self.domain.primary not in self.aspect.bound_spaces:
            self.aspect.bound_spaces[self.domain.primary] = {
                "ub": [],
                "lb": [],
                "eq": [],
            }

        # Sample will figure these out if needed
        _ = self.time
        _ = self.space

    def V(
        self,
        parameter: float | list[float] | None = None,
        length: int | None = None,
    ) -> V:
        """
        Returns a gana variable (V) using .domain as the index.
        If time and space are (or) not given, i.e. .spaced or .timed are False,
        They can be determined.

        For identifying the appropriate temporal scale (index):
            The length of the parameter set is used
            Alternatively, the length of the parameter set can passed directly
            If neither are passed, it defaults to horizon

        If the spatial index is not given, it defaults to the network.

        :param parameters: the parameter/parameter set. Defaults to None.
        :type parameters: int | list, optional
        :param length: length of the parameter set. Defaults to None.
        :type length: int, optional
        :param report: to make a binary reporting variable. Defaults to False.
        :type report: bool, optional
        :param incidental: if this is an incidental calculation. Defaults to False.
        :type incidental: bool, optional

        .. note::
            - parameters and length are mutually exclusive
        """

        self._init_V(parameter, length)

        # with lag it is assumed that the variable of which this is a lagged subset is
        # already defined
        # for example, if opr_t = opr_t-1 + x_t, then opr_t is already defined
        if self.domain.lag:
            return self.Vlag()

        # the reason we check by string is that:
        # some variables can serve as indices, a normal check ends by
        # creating a constraint variable == variable

        # TODO: run this check for all V types.
        # TODO: Will require separate aspect.indices lists
        if str(self.I) not in [str(i) for i in self.aspect.indices]:

            # if a variable has not been created for the self.I
            # create a variable
            # all energia variables are mutable by default
            setattr(
                self.program,
                self.aspect.name,
                V(*self.I, mutable=True, ltx=self.aspect.latex),
            )
            self._inform()

        var = getattr(self.program, self.aspect.name)(*self.I)
        return var

    def Vinc(self, parameters: float | list = None, length: int = None) -> V:
        """
        Returns the incidental variable

        :param parameters: the parameter/parameter set. Defaults to None.
        :type parameters: float | list, optional
        :param length: length of the parameter set. Defaults to None.
        :type length: int, optional

        :returns: the incidental variable
        :rtype: V
        """
        self._init_V(parameters, length)

        if self.domain.lag:
            return self.Vlag()

        self.hasinc = True

        # if incidental calculation is needed
        # incidental calculations do not scale with variable value
        # rather, they are incurred if the reporting binary = 1
        # see the equations below:
        #   calc_total = calc + calc_incidental
        #   calc = v * param
        #   calc_incidental = v_reporting * param_incidental
        # named with a superscript inc
        if self.aspect.latex:
            ltx = self.aspect.latex + r"^{inc}"
        else:
            ltx = self.aspect.name + r"^{inc}"

        # create an incidental variable (continuous)
        setattr(
            self.program,
            f"{self.aspect.name}_incidental",
            V(*self.I, mutable=True, ltx=ltx),
        )
        return getattr(self.program, f"{self.aspect.name}_incidental")(*self.I)

    def Vb(self) -> V:
        r"""
        Bound Variable
        These apply when there are multiple levels of variable-making
        Endogenous bounds apply, i.e.

        :returns: the bound variable
        :rtype: V

        .. math::
           \mathbf{v}_{\dots, t^{+}} <= {\theta}_{\dots, t^{+}} \cdot \mathbf{v}_{\dots, t^{-}}

        where :
        - :math:`\mathbf{v}` is the variable
        - :math:`\theta` is the parameter set
        - :math:`t^{i}` are bespoke discretizations of the horizon
        - :math:`i \in \{-, +\}` are variable-making levels
        """

        bound_aspect = getattr(self.model, self.aspect.bound)

        if bound_aspect not in self.model.dispositions:
            return 1

        if (
            self.domain.space
            not in self.model.dispositions[bound_aspect][self.domain.primary]
        ):

            # if the bound variable has not been defined at the given space
            logger.info(
                "Aspect (%s) not defined at %s, a variable will be created assuming %s as the temporal index",
                bound_aspect,
                self.domain.space,
                self.model.horizon,
            )

            domain = self.domain.change({"periods": self.model.horizon})

        else:
            # if the bound variable has been defined for the given space
            times = self.model.dispositions[bound_aspect][self.domain.primary][
                self.domain.space
            ]
            time = max(list(times))
            if time >= self.domain.periods:
                domain = self.domain.change({"periods": time})
            else:
                # this is if the binding variable has a sparser temporal index compared to time
                raise ValueError(
                    f"Incompatible temporal indices: {bound_aspect} defined in time {time} and {self.aspect} defined in time {self.domain.periods}\n"
                    f"Binding variable ({bound_aspect}) cannot have a denser discretization than variable being bound ({self.aspect})",
                )
        return bound_aspect(domain=domain).V()

    def X(self, parameters: float | list = None, length: int = None) -> V:
        r"""
        Binary Reporting Variable
        These report whether a variable has been made or not
        Also useful to make the variable space semi-continuous

        .. math::
              \mathbf{v}_{\dots, t} <= {\theta}_{\dots, t} \cdot \mathbf{x}_{\dots, t}

        where :
            - :math:`\mathbf{v}` is the variable
            - :math:`\mathbf{x}` is the binary reporting variable
            - :math:`\theta` is the parameter set
            - :math:`t` some bespoke discretization of the horizon
        """

        self._init_V(parameters, length)

        if self.domain.lag:
            return self.Vlag()

        # these are basically named using a breve over the variable name or latex name

        if self.aspect.latex:
            ltx = r"{\breve{" + self.aspect.latex + r"}}"
        else:
            ltx = r"{\breve{" + self.aspect.name + r"}}"
        # create a binary variable
        setattr(
            self.program,
            f"x_{self.aspect.name}",
            V(
                *self.I,
                mutable=True,
                ltx=ltx,
                bnr=True,
            ),
        )
        v_rpt = getattr(self.program, f"x_{self.aspect.name}")
        self.aspect.reporting = v_rpt
        return v_rpt(*self.I)

    def obj(self, maximize: bool = False):
        """
        Set the sample itself as the objective

        :param max: if maximization, defaults to False
        :type max: bool, optional
        """
        if not self.timed:
            # if the temporal index is not passed
            self.domain.periods = self.model.horizon
        if not self.spaced:
            # if the spatial index is not passed
            self.domain.location = self.model.network

        # consider all of self.domain
        v = self.V()

        if len(v) == 1:
            _obj = v
        else:

            _obj = sigma(v)

        if self.hasinc:
            # if there is an incidental variable
            # the incidental variable is added to the objective
            v_inc = self.Vinc()
            if len(v_inc) == 1:
                _obj += v_inc
            else:
                _obj += sigma(v_inc)

        if maximize:
            setattr(self.program, f"max{self.aspect.name})", sup(_obj))
        else:
            setattr(self.program, f"min({self.aspect.name})", inf(_obj))

        self.program.renumber()

    def opt(self, maximize: bool = False):
        """
        Optimize

        :param max: if maximization, defaults to False
        :type max: bool, optional
        """
        self.obj(maximize)
        # optimize!
        self.program.opt()

    def bounds(self):
        """Finds the bounds of the variable"""
        # TODO
        self.opt()
        bmin = self.program.obj()
        self.opt(maximize=True)
        bmax = -self.program.obj()
        return (bmin, bmax)

    def prep(self, nominal: float = 1, norm: bool = True) -> Self:
        """
        Nominal value

        :param nominal: If the input argument (bounds) are to be scaled, defaults to 1
        :type nominal: float, optional
        :param norm: If the input argument (bounds) are normalized, defaults to True
        :type norm: bool, optional
        """
        self.nominal = nominal
        self.norm = norm
        return self

    def output(self, aslist: bool = False, asdict: bool = False, compare: bool = False):
        """
        Solution

        :param aslist: Returns the solution as list, otherwise as a variable
        :type aslist: bool, optional
        :param asdict: Returns the solution as dict, otherwise as a variable
        :type asdict: bool, optional
        :param compare: If True, compares the solutions across multiple solves
        :type compare: bool, optional
        """
        return self.V().output(aslist=aslist, asdict=asdict, compare=compare)

    def eval(self, *values: float):
        """
        Evaluate the variable using parametric variable values

        :param values: values for the parametric variables
        :type values: float
        """
        return self.V().eval(*values)

    def forall(self, index) -> Self:
        """Returns the function at the given index"""
        self._forall = index
        return self

    def __getattr__(self, other):
        aspect = getattr(self.model, other)
        return aspect(self)

    def __le__(self, other):

        Bind(sample=self, parameter=other, leq=True, forall=self._forall)

    def __ge__(self, other):
        Bind(sample=self, parameter=other, geq=True, forall=self._forall)

    def __eq__(self, other):

        if other is True:
            # if a truth value is give
            # just declare the variable
            # it will be non-negative by default
            # and will begin a commodity balance
            self.V()

        elif isinstance(other, Sample):
            if self.aspect.name == other.aspect.name:
                # if self.domain == other.domain:
                return True
            return False

        else:
            if callable(self.of):
                _ = self.of(*self.domain.index_primary[1:]) == True
            Bind(sample=self, parameter=other, eq=True, forall=self._forall)

    def __gt__(self, other):
        logger.info(
            "Bind %s > %s is being written as %s >= %s", self, other, self, other
        )
        _ = self >= other

    def __lt__(self, other):
        logger.info(
            "Bind %s < %s is being written as %s <= %s", self, other, self, other
        )
        _ = self <= other

    def __call__(self, *index) -> Sample:
        return self.aspect(*{*self.domain.index_short, *index}, report=self.report)

    def __getitem__(self, calculate: Sample):
        if isinstance(calculate, int):
            f = self.F(self.F.index[calculate])
            f.report = self.report
            return f
        return calculate(self(), *self.domain.index_spatiotemporal)

    def draw(self, **kwargs):
        """Draws the variable"""
        v = self.V()
        v.draw(**kwargs)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    # This is for binding sample operations, eg. sample + sample or sample - sample

    def __add__(self, other: Self | FuncOfSamples):
        if isinstance(other, (int, float)):
            return FuncOfSamples(F=self.F + other, program=self.program)
        return FuncOfSamples(F=self.F + other.F, program=self.program)

    def __radd__(self, other):
        if not other:
            return self

    def __sub__(self, other: Self | FuncOfSamples):
        if isinstance(other, (int, float)):
            return FuncOfSamples(F=self.F - other, program=self.program)
        return FuncOfSamples(F=self.F - other.F, program=self.program)

    def __rsub__(self, other: int | float):
        return FuncOfSamples(F=other - self.F, program=self.program)

    def __mul__(self, other: Self | FuncOfSamples):
        return FuncOfSamples(F=self.F * other.F, program=self.program)

    def __rmul__(self, other: int | float):
        return FuncOfSamples(F=other * self.F, program=self.program)


class FuncOfSamples:
    """Some Function of Samples

    This is used to bind a function of variables to a given parameter (set)

    """

    def __init__(self, F: F, program: Prg):
        self.program = program
        self.F = F

    def __add__(self, other: Self | Sample):
        if isinstance(other, (int, float)):
            return FuncOfSamples(F=self.F + other, program=self.program)
        return FuncOfSamples(F=self.F + other.F, program=self.program)

    def __radd__(self, other):
        if not other:
            return self

    def __sub__(self, other: Self | Sample):
        if isinstance(other, (int, float)):
            return FuncOfSamples(F=self.F - other, program=self.program)
        return FuncOfSamples(F=self.F - other.F, program=self.program)

    def __rsub__(self, other: int | float):
        return FuncOfSamples(F=other - self.F, program=self.program)

    def __mul__(self, other: Self | Sample):
        return FuncOfSamples(F=self.F * other.F, program=self.program)

    def __rmul__(self, other: int | float):
        return FuncOfSamples(F=other * self.F, program=self.program)

    def __eq__(self, other):
        func = self.F == other
        setattr(self.program, f"eq_{self.F.name}", func)
        return func

    def __le__(self, other):
        func = self.F <= other
        setattr(self.program, f"le_{self.F.name}", func)
        return func

    def __ge__(self, other):
        func = self.F >= other
        setattr(self.program, f"ge_{self.F.name}", func)
        return func

    def opt(self, maximize=False):
        """Optimize the function

        :param max: if maximization, defaults to False
        :type max: bool, optional
        """

        setattr(self.program, f"min_{self.F.name}", inf(self.F))
        self.program.opt()



"""Conversion"""

from __future__ import annotations

from collections.abc import Mapping
from functools import cached_property
from typing import TYPE_CHECKING, Self

from ..._core._hash import _Hash
from ...components.temporal.lag import Lag
from ...components.temporal.modes import Modes

if TYPE_CHECKING:
    from gana import Prg

    from ...components.commodities.commodity import Commodity
    from ...components.operations.operation import Operation
    from ...components.operations.storage import Storage
    from ...components.spatial.linkage import Linkage
    from ...components.spatial.location import Location
    from ...components.temporal.periods import Periods
    from ...represent.model import Model
    from ..indices.sample import Sample


class Conversion(Mapping, _Hash):
    """
    Processes convert one Commodity to another Commodity
    Conversion provides the conversion of resources

    :param resource: Commodity that is balanced
    :type resource: Commodity
    :param operation: Process or Storage that serves as primary spatial index
    :type operation: Process | Storage
    :param bind: Sample that is used to define the conversion
    :type bind: Sample

    :ivar name: Name of the component, generated based on the operation.
    :vartype name: str
    :ivar base: Basis Commodity, usually has a value 1 in the conversion matrix, set using __call__. Defaults to None.
    :vartype base: Commodity
    :ivar conversion: {Commoditys: conversion}. Defaults to {}, set using __eq__.
    :vartype conversion: dict[Commodity : int | float]
    :ivar lag: Temporal lag (processing time) for conversion, set using __getitem__.
    :vartype lag: Lag
    :ivar periods: Periods over which the conversion is defined. Defaults to None.
    :vartype periods: Periods

    :raises ValueError: If conversion lists are of inconsistent lengths.

    .. note::
        - name and operation are generated post init
        - base (__call__), conversion (__eq__), lag (__getitem__) are defined as the program is built
        - Storage contains two processes (charge and discharge), hence is provided separately

    """

    def __init__(
        self,
        aspect: str = "",
        add: str = "",
        sub: str = "",
        operation: Operation | Storage | None = None,
        resource: Commodity | None = None,
        balance: dict[Commodity, float | list[float]] | None = None,
        hold: int | float | None = None,
        attr_name: str = "",
        symbol: str = "η",
        use_max_time: bool = False,
    ):

        self.resource = resource
        self.operation = operation

        self.symbol = symbol

        # * Aspect that elicits the conversion
        self.aspect = aspect

        # * Aspects corresponding to positive and negative conversion
        self.add = add
        self.sub = sub

        if balance:
            self.balance = balance
        else:
            self.balance = {}

        # value to hold, will be applied later
        # occurs when Conversion/Commodity == parameter is used
        # the parameter is held until a dummy resource is created
        self.hold = hold
        # used if a resource is expected to be inventoried
        self.expect: Commodity | None = None

        self.lag: Lag | None = None

        # this is carried forth incase, piece wise linear conversion is used
        self.attr_name = attr_name

        self.use_max_time = use_max_time

    @property
    def args(self) -> dict[str, str | Operation | Commodity | None]:
        """Arguments of the conversion"""
        return {
            'by': self.aspect,
            'add': self.add,
            'sub': self.sub,
            'operation': self.operation,
            'basis': self.resource,
        }

    @classmethod
    def from_balance(
        cls,
        balance: dict[Commodity, float | list[float]],
        by: str = "",
        add: str = "",
        sub: str = "",
        operation: Operation | None = None,
        basis: Commodity | None = None,
    ) -> Self:
        """Creates Conversion from balance dict"""
        conv = cls()
        # set first resource as the basis
        conv.balance = balance
        conv.operation = operation
        conv.aspect = by
        conv.add = add
        conv.sub = sub
        conv.resource = basis
        return conv

    @property
    def name(self) -> str:
        """Name"""
        if self.resource:
            return f"{self.symbol}({self.operation}, {self.resource})"
        return f"{self.symbol}({self.operation})"

    @cached_property
    def model(self) -> Model | None:
        """energia Model"""
        return next((i.model for i in self.balance), None)

    @cached_property
    def program(self) -> Prg | None:
        """gana Program"""
        return self.operation.program

    def balancer(self):
        """
        Checks if there is a list in the conversion
        If yes, tries to make everything consistent
        """

        def _balancer(conversion: dict):

            # check if lists are provided
            check_list = dict.fromkeys(conversion.keys(), False)
            # check lengths of the list, for parameter the length is 1
            check_len = dict.fromkeys(conversion.keys(), 1)

            for res, par in conversion.items():
                if isinstance(par, list):
                    check_list[res] = True
                    check_len[res] = len(par)

            # check if all the list lens are the same
            lengths = {i for i in check_len.values() if i > 1}

            if len(lengths) > 1:
                # if there are different lengths, raise an error
                raise ValueError(
                    f"Conversion: {self.name} has inconsistent list lengths: {lengths}",
                )

            if any(check_list.values()):
                length = next(iter(lengths))
                # if any of the values are a list
                #
                for res, par in conversion.items():
                    if isinstance(par, (float, int)):
                        conversion[res] = [par] * length
            return conversion

        self.balance = _balancer(self.balance)

    def write(
        self, space: Location | Linkage, time: Periods | Lag, modes: Modes | None = None
    ):
        """Writes equations for conversion balance"""

        def time_checker(res: Commodity, space: Location | Linkage, time: Periods):
            """This checks if it is actually necessary
            to write conversion at denser temporal scales
            """
            # This checks whether some other aspect is defined at
            # a lower temporal scale

            if space not in self.model.balances[res]:
                # if not defined for that location, check for a lower order location
                # i.e. location at a lower hierarchy,
                # e.g. say if space being passed is a city, and a grb has not been defined for it
                # then we need to check at a higher order
                parent = self.model.space.split(space)[
                    1
                ]  # get location at one hierarchy above
                if parent:
                    # if that indeed exists, then make the parent the space
                    # the conversion Balance variables will feature in grb for parent location
                    space = parent

            _ = self.model.balances[res][space][time]

            if res.inv_of:
                # for inventoried resources, the conversion is written
                # using the time of the base resource's grb
                res = res.inv_of

            try:
                times = list(
                    [
                        t
                        for t in self.model.balances[res][space]
                        if self.model.balances[res][space][t]
                    ],
                )
            except KeyError:
                times = []
            # write the conversion balance at
            # densest temporal scale in that space
            if times:
                if self.use_max_time:
                    return max(times)
                return min(times)

            return time.horizon

        for res, par in self.items():

            if res in self.model.balances:
                time = time_checker(res, space, time)
                _ = self.model.balances[res].get(space, {})

            eff = par if isinstance(par, list) else [par]

            decision = getattr(self.operation, self.aspect)

            if eff[0] < 0:
                # Resources are consumed (expendend by Process) immediately

                dependent = getattr(res, self.sub)
                eff = [-e for e in eff]
            else:
                # Production — may occur after lag
                time = self.lag.of if self.lag else time
                dependent = getattr(res, self.add)

            if modes:
                rhs = dependent(decision, space, modes, time)

                lhs = decision(space, modes, time)
            else:
                rhs = dependent(decision, space, time)

                lhs = decision(space, time)

            _ = lhs[rhs] == eff

    def items(self):
        """Items of the conversion balance"""
        return self.balance.items()

    def keys(self):
        """Keys of the conversion balance"""
        return self.balance.keys()

    def values(self):
        """Values of the conversion balance"""
        return self.balance.values()

    def __getitem__(self, key: Commodity) -> float | list[float]:
        """Used to define mode based conversions"""
        return self.balance[key]

    def __setitem__(self, key: Commodity, value: float | list[float]):
        self.balance[key] = value

    def __call__(self, basis: Commodity | Conversion, lag: Lag | None = None) -> Self:
        # sets the basis
        if isinstance(basis, Conversion):
            # if a Conversion is provided (parameter*Commodity)
            # In this case the associated conversion is not 1
            # especially useful if Process is scaled to consumption of a commodity
            # i.e. basis = -1*Commodity
            self.balance = {**self, **basis}
            self.resource = next(iter(self))

        else:
            # if a Commodity is provided
            # implies that the conversion is 1
            # i.e the Process is scaled to one unit of this Commodity produced
            self.balance = {basis: 1.0, **self}

        if lag:
            self.lag = lag

        return self

    def __eq__(
        self,
        other: Conversion | list[Conversion] | int | float | dict[Modes, Conversion],
    ):
        if isinstance(other, dict):

            collect_parents = []
            for mode, conv in other.items():
                conv.operation = self.operation

                collect_parents.append(mode.parent)

                other[mode] = Conversion.from_balance({**self, **conv}, **self.args)

            if len(set(collect_parents)) > 1:
                raise ValueError(
                    f"{self}: PWL Conversion modes must belong to the same parent Modes",
                )

            if len(collect_parents[0]) != len(other):
                raise ValueError(
                    f"{self}: PWL Conversion modes must account for all modes in {collect_parents[0]}",
                )

            setattr(
                self.operation,
                self.attr_name,
                PWLConversion.from_balance(
                    balance=other, sample=getattr(self.operation, self.aspect)
                ),
            )
            return getattr(self.operation, self.attr_name)

        if isinstance(other, list):
            if self.resource:
                for i, o in enumerate(other):
                    other[i] = Conversion.from_balance(
                        {**self, **o},
                    )
            setattr(
                self.operation,
                self.attr_name,
                PWLConversion(
                    conversions=other,
                    sample=(
                        getattr(self.operation.stored, self.aspect)
                        if hasattr(self.operation, "stored")
                        else getattr(self.operation, self.aspect)
                    ),
                    aspect=self.aspect,
                    add=self.add,
                    sub=self.sub,
                ),
            )
            return getattr(self.operation, self.attr_name)

        if isinstance(other, (int, float)):
            # this is used for inventory conversion
            # when not other resource besides the one being inventoried is involved
            self.balance = {**self, self.expect: 1 / -float(other)}
        else:
            self.balance = {**self, **other}
        self.model.convmatrix[self.operation] = self.balance
        return self

    def __neg__(self) -> Self:
        self.balance = {res: -par for res, par in self.balance.items()}
        return self

    def __add__(self, other: Conversion) -> Self:
        self.balance = {**self, **other}
        return self

    def __sub__(self, other: Conversion) -> Self:
        self.balance = {**self, **-other}
        return self

    def __mul__(self, times: int | float | list) -> Self:
        if isinstance(times, list):
            self.balance = {res: [par * i for i in times] for res, par in self.items()}
        else:
            self.balance = {res: par * times for res, par in self.items()}
        return self

    def __rmul__(self, times) -> Self:
        return self * times

    def __len__(self):
        """Length of the conversion balance"""
        return len(self.balance)

    def __iter__(self):
        return iter(self.balance)


class PWLConversion(Mapping, _Hash):
    """Piece Wise Linear Conversion"""

    def __init__(
        self,
        conversions: list[Conversion],
        sample: Sample,
        aspect: str = "",
        add: str = "",
        sub: str = "",
    ):

        self.sample = sample
        self.operation = self.sample.domain.operation or self.sample.domain.primary
        self.model = self.sample.model
        if conversions:

            self.modes = self.model.Modes(size=len(conversions), sample=sample)
            self.balance: dict[Modes, Conversion] = dict(zip(self.modes, conversions))

            for conv in conversions:
                conv.operation = self.operation
                conv.aspect = aspect or conv.aspect
                conv.add = add or conv.add
                conv.sub = sub or conv.sub

        self._aspect = aspect
        self._add = add
        self._sub = sub

    @property
    def aspect(self) -> str:
        if self._aspect:
            return self._aspect
        return self[0].aspect

    @property
    def add(self) -> str:
        if self._add:
            return self._add
        return self[0].add

    @property
    def sub(self) -> str:
        if self._sub:
            return self._sub
        return self[0].sub

    @property
    def lag(self) -> str:
        return self[0].lag

    @classmethod
    def from_balance(
        cls,
        balance: dict[Modes, Conversion],
        sample: Sample,
    ) -> Self:
        """Creates PWLConversion from balance dict"""
        conv = cls([], sample)
        conv.operation = sample.domain.operation
        conv.balance = balance
        conv.modes = (next(iter(balance))).parent

        return conv

    def balancer(self):
        """Balances all conversions"""
        for conv in self.balance.values():
            conv.balancer()

    @property
    def name(self) -> str:
        """Name"""
        return f"η_PWL({self.operation}, {self.modes})"

    def __len__(self):
        """Length of the conversion balance"""
        return len(self.balance)

    def __iter__(self):
        return iter(self.balance)

    def __setitem__(self, key: int | str | Modes, value: Conversion):

        if isinstance(key, int):
            key = self.modes[key]

        self.balance[key] = value

    def __getitem__(self, key: int | str) -> Conversion:
        """Used to define mode based conversions"""

        if isinstance(key, int):
            key = self.modes[key]

        return self.balance[key]

    def items(self):
        """Items of the conversion balance"""
        return self.balance.items()

    def keys(self):
        """Keys of the conversion balance"""
        return self.balance.keys()

    def values(self):
        """Values of the conversion balance"""
        return self.balance.values()

    def box(self):
        """Consolidates the conversion dict into {resource: par} format"""
        resources = list(set().union(*self.values()))
        _box = {r: [] for r in resources}
        for conv in self.values():
            for resource in resources:
                if resource in conv:
                    _box[resource].append(conv[resource])
                else:
                    _box[resource].append(None)

        return _box

    def write(self, space: Location | Linkage, time: Periods | Lag):
        """Writes equations for conversion balance"""

        for mode, conv in self.items():
            conv.write(space, time, mode)



"""Inherited _Component class"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from gana import I

from ._x import _X

if TYPE_CHECKING:
    from ..components.spatial.location import Location
    from ..components.temporal.periods import Periods
    from ..dimensions.problem import Problem
    from ..dimensions.space import Space
    from ..dimensions.time import Time
    from ..modeling.indices.sample import Sample
    from ..represent.model import Model


class _Component(_X):
    """
    A component with a mathematical program (hence index in it) and
    input parameters (modeling aspects).

    :param label: Label for the component. Defaults to None.
    :type label: str, optional
    :param citations: Citation for the component. Defaults to None.
    :type citations: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]

    :raises AttributeError: If an invalid parameter is provided for the component type.

    .. note:
        - `name`, `model` set when component assigned to Model attribute.
        - `constraints` and `domains` are populated as the program is built.
    """

    def __init__(
        self,
        label: str = "",
        citations: str = "",
        **kwargs,
    ):
        self.parameters = kwargs

        # what differentiates a component from an index is that it has aspects
        # that we can control to adjust their states of existence
        _X.__init__(self, label=label, citations=citations)

    @property
    def network(self) -> Location:
        """Circumscribing Loc (Spatial Scale)"""
        return self.model.network

    @property
    def horizon(self) -> Periods:
        """Circumscribing Periods (Temporal Scale)"""
        return self.model.horizon

    @cached_property
    def time(self) -> Time:
        """Time"""
        return self.model.time

    @cached_property
    def space(self) -> Space:
        """Space"""
        return self.model.space

    @cached_property
    def problem(self) -> Problem:
        """Feasible Region"""
        # the decision tree gives the component access to
        # the aspects of all other components
        return self.model.problem

    def _handle_norm(self, aspect: str, sample: Sample) -> Sample:
        """Check if a nominal parameter is set for the aspect"""
        # check if a request to normalize has been made
        nominal = aspect + "_nominal"
        nom = aspect + "_nom"
        normalize = aspect + "_normalize"
        norm = aspect + "_norm"

        if normalize in self.parameters or norm in self.parameters:
            _normalize = self.parameters.get(normalize, self.parameters.get(norm))
        else:
            # default behavior is to normalize
            _normalize = False

        # irrespective of normalize request, check if nominal value set
        if nominal in self.parameters or nom in self.parameters:
            # if nominal is given, normalize to it
            return sample.prep(self.parameters[nominal], True)

        if _normalize:
            # if _normalize is True but no nominal provided
            return sample.prep(norm=True)

        return sample

    def _handle_x(self, aspect: str, sample: Sample) -> Sample:
        """Check if aspect is optional"""

        if (
            aspect + '_optional' in self.parameters
            and self.parameters[aspect + '_optional']
        ):
            return sample.x
        if (
            aspect + '_report' in self.parameters
            and self.parameters[aspect + '_report']
        ):
            return sample.x

        return sample

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

        # this handles the parameters being set on init
        if name == "model" and value is not None:

            model: Model = value

            for attr, param in self.parameters.items():

                if attr in model.manual:
                    _ = getattr(self, attr) == param
                    continue

                # attributes name expected here are of the format <aspect>_<bound>
                # for exact equality, just <aspect>

                split_attr = attr.split('_')
                # irrespective of exact or bound, first part is always the aspect
                aspect = split_attr[0]

                if isinstance(self, getattr(self.model, aspect).primary_type):
                    # check type match

                    # get the sample
                    sample = getattr(self, aspect)

                    if isinstance(param, list):
                        sample = self._handle_x(
                            aspect, self._handle_norm(aspect, sample)
                        )

                    else:
                        sample = self._handle_x(aspect, sample)

                    if len(split_attr) == 1:
                        # if split returned just the aspect name
                        # then it's an equality
                        _ = sample == param

                    # else, check if lower or upper bound

                    elif split_attr[1] in ["max", "ub", "UB", "leq"]:

                        _ = sample <= param

                    elif split_attr[1] in ["min", "lb", "LB", "geq"]:
                        _ = sample >= param

                else:
                    # error if type mismatch
                    raise AttributeError(
                        f"Parameter {attr} valid for {self.model.cookbook[aspect].primary_type} not {type(self).__name__}"
                    )

    def __getattr__(self, name):

        if self.model:
            # no need to run a hasattr check
            # let it raise an attribute error if not found
            aspect = getattr(self.model, name)
            if callable(aspect):
                return aspect(self)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'",
        )

    @cached_property
    def I(self):
        """gana index set (I)"""
        _index = I(self.name, tag=self.label or "")
        setattr(self.program, self.name, _index)
        return _index

    def __len__(self):
        """Length of the component's index set"""
        # always one
        return 1



"""Operational Mode attached to a Parameter"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Self

from gana import I

from ..._core._x import _X

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ...modeling.indices.sample import Sample


class Modes(_X):
    """
    Represents a discrete choice to be taken within a spatiotemporal disposition.
    Modes can split usage. A Mode of Operation can be used for Conversion, Use, etc.

    :param n_modes: Number of modes. Defaults to 1.
    :type n_modes: int
    :param sample: The aspect and component (sample) which is being 'moded'. Defaults to None.
    :type sample: Sample | None
    :param parent: Parent mode, if any. Defaults to None.
    :type parent: Self | None
    :param pos: Position in the parent mode set. Defaults to None.
    :type pos: int | None

    :ivar name: The name of the mode, usually a number.
    :vartype name: str | float | int
    """

    def __init__(
        self,
        size: int = 1,
        sample: Sample | None = None,
        parent: Self | None = None,
        n: int | None = None,
        label: str = "",
        citations: str = "",
    ):
        self.size = size
        self.sample = sample
        self.parent = parent
        self.n = n

        _X.__init__(self, label=label, citations=citations)

        if self.n is not None:
            self.name = f"{self.parent}[{self.n}]"

        if self.parent:
            self.model = self.parent.model

        # have child modes been made
        self._birthed = False
        # where child modes are stored
        self._ = []

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        # overwrite X.cons property
        # this gets the actual constraint objects from the program
        # based on the pname (attribute name) in the program
        if self.parent:
            return [getattr(self.program, c) for c in self.constraints]
        return list(
            set(
                [getattr(self.program, c) for c in self.constraints]
                + sum(
                    [m.cons for m in self],
                    [],
                )
            )
        )

    @cached_property
    def I(self) -> I:
        """Index set of modes"""

        if self.parent:
            _ = self.parent.I  # makes sure the parent is indexed
            # do not set a new index set, get from parent
            return getattr(self.parent.program, self.parent.name)[self.n]

        _index = I(size=self.size, tag=f"Modes of {self.sample.aspect}")

        setattr(
            self.program,
            self.name,
            _index,
        )

        return _index

    def __eq__(self, other: Modes) -> bool:
        """Equality operator"""
        if isinstance(other, Modes):
            if other.name == self.name:
                return True
        return False

    def __len__(self) -> int:
        """Length of the modes"""
        return self.size

    def __getitem__(self, key: int) -> _X:
        """Get a mode by index"""
        if not self._birthed:
            self._ = [
                Modes(sample=self.sample, parent=self, n=i) for i in range(self.size)
            ]
            self._birthed = True

        return self._[key]

    def __iter__(self):
        """Iterate over modes"""
        for i in range(self.size):
            yield self[i]



"""Land"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .commodity import Commodity

if TYPE_CHECKING:
    from ..measure.unit import Unit


class Land(Commodity):
    """
    Land used by Operations

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """

    def __init__(
        self,
        basis: Unit | None = None,
        label: str = "",
        citations: str = "",
        **kwargs,
    ):
        Commodity.__init__(
            self, basis=basis, label=label, citations=citations, **kwargs
        )



"""Variable"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Self, Type

import matplotlib.pyplot as plt
from matplotlib import rc

from ...components.commodities.commodity import Commodity
from ...components.game.couple import Interact
from ...components.game.player import Player
from ...components.impact.indicator import Indicator
from ...components.operations.process import Process
from ...components.operations.storage import Storage
from ...components.operations.transport import Transport
from ...components.spatial.linkage import Linkage
from ...components.spatial.location import Location
from ...components.temporal.lag import Lag
from ...components.temporal.modes import Modes
from ...components.temporal.periods import Periods
from ...dimensions.space import Space
from ...dimensions.time import Time
from ...utils.dictionary import merge_tree_levels
from ..constraints.balance import Balance as BalCons
from ..constraints.vmap import Map as MapCons
from ..indices.domain import Domain
from ..indices.sample import Sample

if TYPE_CHECKING:
    from gana import I as Idx
    from gana import Prg
    from gana import V as Var
    from gana.sets.constraint import C

    from ..._core._component import _Component
    from ..._core._x import _X
    from ...dimensions.problem import Problem
    from ...represent.model import Model


@dataclass
class Aspect:
    r"""
    A particular facet of the system under consideration. A sample of an aspect at a
    specific disposition is represented by a variable (:math:`\overset{\*}{v} \in \overset{\*}{\mathcal{V}}`).
    The range of values which the :math:`\overset{\*}{v}` is bounded such that:

    .. math::

        \overset{\*}{v} \in [\underline{\theta}, \overline{\theta}]

    Aspects can be decision-variables (:math:`\dot{v}`) that prescribe a control action
    or a set point for the state, or derived (calculated) variables (:math:`\hat{v}`) that
    represent the state of the system. States can be operation capacity, production levels,
    purchase levels, emissions, consumption levels, etc.


    :param primary_types: associated components type(s). Defaults to None.
    :type primary_types: Type[_Component] | tuple[Type[_Component], ...]
    :param nn: If True, the decision is a non-negative decision. Defaults to True.
    :type nn: bool
    :param ispos: If True, the decision is positive (non-negative). Defaults to True.
    :type ispos: bool
    :param neg: Negative form or representation of the decision, if any. Defaults to "".
    :type neg: str
    :param latex: LaTeX string. Defaults to "".
    :type latex: str
    :param bound: if the aspect is bounded by another. Defaults to "".
    :type bound: str
    :param label: Label for the decision. Defaults to "".
    :type label: str

    :ivar model: Model to which the Aspect belongs.
    :vartype model: Model
    :ivar name: Name of the Aspect.
    :vartype name: str
    :ivar indices: List of indices (Location, Periods) associated with the Aspect.
    :vartype indices: list[Location | Linkage, Periods]
    :ivar bound_spaces: Spaces where the Aspect has been already bound.
    :vartype bound_spaces: dict[Commodity | Process | Storage | Transport, list[Location | Linkage]]
    :ivar domains: List of domains associated with the Aspect.
    :vartype domains: list[Domain]

    :raises ValueError: If `primary_type` is not defined.
    """

    primary_type: Type[_Component] | tuple[Type[_Component], ...]
    nn: bool = True
    ispos: bool = True
    neg: str = ""
    latex: str = ""
    bound: str = ""
    label: str = ""

    def __post_init__(self):
        # will be set when added to model
        self.name: str = ""

        # name of the decision
        self.model: Model | None = None

        if self.label:
            if self.nn:
                self.label += " [+]"
            else:
                self.label += " [-]"
        self.indices: list[Location | Linkage | Periods] = []

        # # if a decision is bounded by another decision
        # self.bound: Self = None

        # spaces where the aspect has been already bound
        self.bound_spaces: dict[
            Commodity | Process | Storage | Transport,
            list[Location | Linkage],
        ] = {}

        # upper/lower/exact bounds are set on these locations/periods

        self.ubs: dict[Location | Linkage, Periods] = {}
        self.lbs: dict[Location | Linkage, Periods] = {}
        self.eqs: dict[Location | Linkage, Periods] = {}

        # Domains of the decision
        self.domains: list[Domain] = []

        # a dictionary of domains and their maps from higher order domains
        # reporting variable
        self.reporting: Var | None = None

        self.constraints: list[str] = []

        # this keeps track of whether GRB has already been added
        self.balances: dict[tuple[Idx, ...], bool] = {}

    @cached_property
    def maps(self) -> dict[Aspect, dict[str, list[Domain]]]:
        """Maps of the decision"""
        self.model.maps[self] = {
            "time": {},
            "space": {},
            "modes": {},
            "samples": {},
        }
        return self.model.maps[self]

    @cached_property
    def maps_report(self) -> dict[Aspect, dict[str, list[Domain]]]:
        """Maps of the decision"""
        self.model.maps_report[self] = {
            "time": {},
            "space": {},
            "modes": {},
            "samples": {},
        }
        return self.model.maps_report[self]

    # @cached_property
    # def maps(self) -> dict[Aspect, dict[str, list[Domain]]]:
    #     """Maps of the decision"""
    #     self.model.maps[self] = {}
    #     return self.model.maps[self]

    # @cached_property
    # def maps_report(self) -> dict[Aspect, dict[str, list[Domain]]]:
    #     """Maps of the decision"""
    #     self.model.maps_report[self] = {s}
    #     return self.model.maps_report[self]

    @cached_property
    def isneg(self) -> bool:
        """Does this remove from the domain?"""
        return not self.ispos

    @cached_property
    def sign(self) -> float:
        """Gives the multiplier in balances"""
        if self.ispos:
            return 1.0
        else:
            return -1.0

    @cached_property
    def space(self) -> Space:
        """Space"""
        return self.model.space

    @cached_property
    def program(self) -> Prg:
        """Mathematical Program"""
        return self.model.program

    @cached_property
    def problem(self) -> Problem:
        """Tree"""
        return self.model.problem

    @cached_property
    def time(self) -> Time:
        """Time"""
        return self.model.time

    @property
    def I(self) -> Idx:
        """gana index set (I)"""
        return getattr(self.program, self.name)

    @property
    def V(self) -> Var:
        """Variable"""
        return getattr(self.program, self.name)

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        return [getattr(self.program, c) for c in self.constraints]

    @property
    def network(self) -> Location:
        """Circumscribing Location (Spatial Scale)"""
        return self.model.network

    @property
    def horizon(self) -> Periods:
        """Circumscribing Periods (Temporal Scale)"""
        return self.model.horizon

    @property
    def dispositions(self) -> dict[
        Self,
        dict[
            Commodity | Process | Storage | Transport,
            dict[
                Periods | Location | Linkage,
                dict[Location | Periods | Linkage, bool],
            ],
        ],
    ]:
        """Dispositions dict"""
        return self.model.dispositions[self]

    @property
    def sizes(self):
        """dict of domain sizes"""
        _sizes = {}
        for d in self.domains:
            if d.size in _sizes:
                _sizes[d.size].append(d.index)
            else:
                _sizes[d.size] = [d.index]
        return _sizes

    @property
    def box(self):
        """Box of domain indices"""
        return [b.index_short for b in self.domains]

    def crumple_domains(self):
        return merge_tree_levels(self.dispositions)

    def alias(self, *names: str):
        """
        Create aliases for the decision

        :param names: Names of the aliases
        :type names: str
        """
        self.model.alias(*names, of=self.name)

    def update(self, domain: Domain, reporting: bool = False):
        """Each inherited object has their own"""

    def show(self, descriptive=False):
        """Pretty print the component"""
        for c in self.cons:
            c.show(descriptive)

    def output(
        self,
        n_sol: int = 0,
        aslist: bool = False,
        asdict: bool = False,
        compare: bool = False,
    ) -> list[float] | dict[tuple[Idx, ...], float] | None:
        """
        Solution

        :param n_sol: Solution number. Defaults to 0.
        :type n_sol: int, optional
        :param compare: Compares the solution with the previous one. Defaults to False.
        :type compare: bool, optional
        :param asdict (bool, optional): Returns values taken as dict. Defaults to False.
        :type asdict: bool, optional
        :param aslist (bool, optional): Returns values taken as list. Defaults to False.
        :type aslist: bool, optional

        :return: List of values taken by the decision.
        :rtype: list[float] | None
        """
        var: Var = getattr(self.program, self.name)
        return var.output(n_sol, aslist=aslist, asdict=asdict, compare=compare)

    def gettime(self, *index) -> list[Periods]:
        """Finds the sparsest time scale in the domains"""
        ds = [i for i in self.indices if all([x in i for x in index])]
        t = [t for t in ds if isinstance(t, Periods)]
        return t

    def draw(
        self,
        x: _X,
        y: tuple[_X] | _X,
        z: tuple[_X] | _X = None,
        font_size: float = 16,
        fig_size: tuple[float, float] = (12, 6),
        linewidth: float = 0.7,
        color: str = "blue",
        grid_alpha: float = 0.3,
        usetex: bool = True,
    ):
        """Plot the decision"""
        if not isinstance(y, tuple):
            y = (y,)
        if not z:
            index = [i.I for i in y + (x,)]
        elif not isinstance(z, tuple):
            z = (z,)

        if usetex:
            rc(
                "font",
                **{"family": "serif", "serif": ["Computer Modern"], "size": font_size},
            )
            rc("text", usetex=usetex)
        else:
            rc("font", **{"size": font_size})

        _, ax = plt.subplots(figsize=fig_size)

        if not z:
            ax.plot(
                [i.name for i in x.I._],
                self.V(*index).output(True),
                linewidth=linewidth,
                color=color,
            )
        else:
            for z_ in z:
                index = [i.I for i in (z_,) + y + (x,)]
                ax.plot(
                    [i.name for i in x.I._],
                    self.V(*index).output(True),
                    linewidth=linewidth,
                    label=z_.label if z_.label else z_.name,
                )

        label_ = [i.label if i.label else i.name for i in y]
        label_ = str(label_).replace("[", "").replace("]", "").replace("'", "")
        plt.title(f"{self.label} - {label_}")
        plt.ylabel("Values")
        plt.xlabel(f"{x.label or x.name}")
        plt.grid(alpha=grid_alpha)
        if z:
            plt.legend()
        plt.rcdefaults()

    def __neg__(self):
        """Negative Consequence"""

        dscn = type(self)(
            nn=False,
            primary_type=self.primary_type,
        )
        dscn.neg, self.neg = self, dscn
        dscn.ispos = not self.ispos
        return dscn

    def __len__(self):
        return len(self.domains)

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, Aspect):
            return self.name == other.name

    def __call__(
        self, *index: _X, domain: Domain | None = None, report=False
    ) -> Sample:

        if not domain:

            args = {
                "indicator": None,
                "commodity": None,
                "player": None,
                "process": None,
                "storage": None,
                "transport": None,
                "periods": None,
                "couple": None,
                "location": None,
                "linkage": None,
                "lag": None,
                "modes": None,
            }

            type_map = {
                Periods: ("periods", "timed", False),
                Lag: ("lag", "timed", False),
                Location: ("location", "spaced", False),
                Linkage: ("linkage", "spaced", False),
                Process: ("process", None, True),
                Storage: ("storage", None, True),
                Transport: ("transport", None, True),
                Player: ("player", None, False),
                Interact: ("couple", None, False),
                Indicator: ("indicator", None, False),
                Modes: ("modes", None, False),
                Commodity: ("commodity", None, True),
            }

            samples: list[Sample] = []
            timed, spaced = False, False

            for comp in index:
                if isinstance(comp, Sample):
                    samples.append(comp)
                    for b in samples:
                        if b.domain.samples:
                            samples.extend(b.domain.samples)
                    samples = list(set(samples))
                    continue

                for typ, (attr, flag, require_primary) in type_map.items():
                    if isinstance(comp, typ):
                        if require_primary and (
                            not self.primary_type
                            or not isinstance(comp, self.primary_type)
                        ):
                            raise ValueError(
                                f"For component {self} of type {type(self)}: "
                                f"{comp} of type {type(comp)} not recognized as an index",
                            )
                        args[attr] = comp
                        if flag == "timed":
                            timed = True
                        elif flag == "spaced":
                            spaced = True
                        break
                else:
                    raise ValueError(
                        f"For component {self} of type {type(self)}: "
                        f"{comp} of type {type(comp)} not recognized as an index",
                    )

            args = {k: v for k, v in args.items() if v is not None}

            if samples:
                args["samples"] = samples

            domain = Domain(**args)

        else:
            timed = spaced = True

        return Sample(
            aspect=self, domain=domain, timed=timed, spaced=spaced, report=report
        )

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __init_subclass__(cls):
        cls.__repr__ = Aspect.__repr__
        cls.__hash__ = Aspect.__hash__

    def __iter__(self):
        """Iterate over domains"""
        for d in self.domains:
            yield self(domain=d)

    def Map(self, domain: Domain, reporting: bool = False):
        """Map the aspect to the domain"""

        MapCons(aspect=self, domain=domain, reporting=reporting)

    def Balance(self, domain: Domain):
        """Add a general resource balance for the aspect over the domain"""

        BalCons(aspect=self, domain=domain)



"""Function for scaling input data"""

from typing import Literal

from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler, StandardScaler


def scaling(data: DataFrame, how: Literal['max', 'minmax', 'standard']) -> DataFrame:
    """Scales input DataFrame

    :param data: input data
    :type data: DataFrame
    :param how: 'max', 'minmax', 'standard'
    :type how: Literal['max', 'minmax', 'standard']

    :return: scaled DataFrame
    :rtype: DataFrame

    :raises ValueError: if data is not DataFrame
    """
    if not isinstance(data, DataFrame):
        raise ValueError("please provide DataFrame")

    if how == "max":
        data = data / data.max()

    else:
        if how == "standard":
            scaler = StandardScaler()
        if how == "minmax":
            scaler = MinMaxScaler()
        data = DataFrame(scaler.fit_transform(data))

    return data



"""Program"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from gana import I, Prg

if TYPE_CHECKING:
    from ..model import Model


@dataclass
class Program(Prg):
    """The mathematical program of the model

    :ivar players: Players index set
    :vartype players: I
    :ivar periods: Periods index set
    :vartype periods: I
    :ivar locations: Locations index set
    :vartype locations: I
    :ivar linkages: Linkages index set
    :vartype linkages: I
    :ivar sources: Sources index set
    :vartype sources: I
    :ivar sinks: Sinks index set
    :vartype sinks: I
    :ivar spaces: Spaces index set
    :vartype spaces: I
    :ivar currencies: Currencies index set
    :vartype currencies: I
    :ivar envs: Environments index set
    :vartype envs: I
    :ivar socs: Societies index set
    :vartype socs: I
    :ivar ecos: Ecosystems index set
    :vartype ecos: I
    :ivar resources: Resources index set
    :vartype resources: I
    :ivar lands: Lands index set
    :vartype lands: I
    :ivar materials: Materials index set
    :vartype materials: I
    :ivar processes: Processes index set
    :vartype processes: I
    :ivar storages: Storages index set
    :vartype storages: I
    :ivar transits: Transits index set
    :vartype transits: I
    :ivar states: States index set
    :vartype states: I
    :ivar controls: Decisions index set
    :vartype controls: I
    :ivar streams: Streams index set
    :vartype streams: I
    :ivar consequences: Consequences index set
    :vartype consequences: I

    .. note::
        - all the index sets are generated post initialization
    """

    model: Model = None

    def __post_init__(self):
        Prg.__post_init__(self)

        # Component Index Sets
        self.name = f"Program({self.model})"

    def __getattr__(self, item):

        if item in self.model.ancestry:
            index = I(mutable=True)
            setattr(self, item, index)
            return index
        raise AttributeError(
            f"{self} has no '{item}'",
        )



"""Time Periods"""

from __future__ import annotations

import logging
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING, Self

from gana import I as Idx

from ..._core._x import _X
from ...modeling.parameters.value import Value
from ...utils.dictionary import NotFoundError, compare
from .lag import Lag

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ...components.temporal.modes import Modes
    from ...dimensions.time import Time


logger = logging.getLogger("energia")


class Periods(_X):
    """
    A discretization of Time.

    :param periods: Number of periods in Periods. Defaults to 1.
    :type periods: int | float
    :param of: The periods of which this is a multiple. Defaults to None.
    :type of: Periods | Lag, optional
    :param name: Name of the periods. Defaults to None.
    :type name: str, optional
    :param label: Label of the periods. Defaults to None.
    :type label: str, optional

    :ivar model: Model to which the Periods belongs.
    :vartype model: Model
    :ivar time: Time to which the Periods belongs.
    :vartype time: Time
    :ivar horizon: Horizon of the Time.
    :vartype horizon: Periods
    :ivar I: Index set of the Periods.
    :vartype I: I
    :ivar constraints: List of constraints associated with the Periods. Defaults to [].
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the Periods. Defaults to [].
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the Periods.
    :vartype aspects: dict[Aspect, list[Domain]]
    """

    def __init__(
        self,
        size: int | float = 1,
        of: Self | None = None,
        n: int | None = None,
        label: str = "",
        citations: str = "",
    ):
        self.size = size

        if of is not None:
            of.isof.append(self)

        self.of = of

        # is a period of
        self.isof: list[Self] = []

        _X.__init__(self, label=label, citations=citations)

        # if this is a slice of another period
        self.slice: slice | None = None

        # if this is a single period in Periods
        self.n = n

        # if parent is true, this is part of another periods set
        self.parent: Self | None = None

        self.modes: list[Modes] = []

        if self.of is not None and self.size:
            # self.tree = {self.of: self.of.tree}
            self.name = f"{self.size}{self.of}"

        self._howmany: dict[Periods, float] = {}

    def isroot(self):
        """Is used to define another period?"""
        if self.of is None:
            return True

    @property
    def tree(self) -> dict[Self, dict]:
        """Tree representation of the Periods"""
        if self.of is None:
            return {self: {}}

        return {self: self.of.tree}

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        return [getattr(self.program, c) for c in self.constraints]

    @property
    def horizon(self) -> Self:
        """Time Horizon"""
        return self.time.horizon

    @property
    def ishorizon(self) -> bool:
        """Is this the horizon of the model?"""
        return self == self.time.horizon

    @cached_property
    def time(self) -> Time:
        """Time to which the Periods belongs"""
        return self.model.time

    @cached_property
    def I(self) -> Idx:
        """Index tuple"""

        # given that temporal scale is an ordered set and not a self contained set
        # any time periods will be a fraction of the horizon
        if self.slice is not None:
            return self.of.I[self.slice]

        if self.n is not None:
            return self.of.I[self.n]

        if self.isof:
            return self.isof[0].I + (self.i,)
        return (self.i,)

    @cached_property
    def i(self) -> Idx:
        """Only Index"""
        _i = Idx(size=self.isof[0].size if self.isof else 1, tag=self.label or "")
        setattr(self.program, self.name, _i)
        return _i

    @property
    def true_size(self) -> float:
        """True size of the Periods"""
        if self.of is None:
            return self.size

        return self.size * self.of.true_size

    def howmany(self, of: Periods):
        """How many periods make this period"""
        try:
            return self._howmany[of]
        except KeyError:
            try:
                _return = compare(self.tree, of)
            except NotFoundError:
                try:
                    _return = 1 / compare(of.tree, self)
                except NotFoundError:
                    try:
                        _return = self.size / compare(of.tree, self.of)
                    except NotFoundError:
                        try:
                            _return = compare(self.tree, of.of) / of.size
                        except NotFoundError:
                            raise ValueError(f"No common basis between {self} and {of}")

        self._howmany[of] = _return
        return _return

    def __mul__(self, times: int | float):

        if times == 0:
            return 0

        if times < 0:
            # if multiplying by a negative number
            # return a lag
            # if not self._of:
            return Lag(of=self, periods=-times)

            # raise ValueError(f"{self} is not a period of anything, so cannot be lagged")

        # if times < 1:
        #     return (1 / times) * self

        return Periods(size=times, of=self)

    def __truediv__(self, other: int | float):
        if isinstance(other, Periods):
            return self.howmany(other)

        return self * (1 / other)

        # return Periods(periods=other, of=self)

    def __rtruediv__(self, other: int | float):
        return Value(value=other, periods=self)

    def __rmul__(self, other: int | float):
        return self * other

    def __call__(self, times):
        return Periods(size=times, of=self)

    def __neg__(self):
        return self * -1

    def __len__(self):
        return int(self.howmany(self.time.horizon))

    def __eq__(self, other: Self | Lag):
        if isinstance(other, Lag):
            return is_(self, other.of)

        return is_(self, other)

    def __ge__(self, other: Self):
        if isinstance(other, Periods):
            return self.time.horizon.howmany(self) <= self.time.horizon.howmany(other)
        raise NotImplementedError

    def __gt__(self, other: Self):
        if isinstance(other, Periods):
            return self.time.horizon.howmany(self) < self.time.horizon.howmany(other)
        raise NotImplementedError

    def __le__(self, other: Self):
        if isinstance(other, Periods):
            return self.time.horizon.howmany(self) >= self.time.horizon.howmany(other)
        raise NotImplementedError

    def __lt__(self, other: Self):
        if isinstance(other, Periods):
            return self.time.horizon.howmany(self) > self.time.horizon.howmany(other)
        raise NotImplementedError

    def __getitem__(self, key: int | slice):

        periods = Periods()
        periods.parent = self
        periods.of = self.of
        periods.model = self.model

        if isinstance(key, slice):
            periods.slice = key
            # record the range
            start = key.start or 0
            stop = key.stop or len(self)
            # not set on the model
            periods.name = rf"{self}[{start}: {stop}]"
        else:
            # single period
            periods.n = key
            periods.name = rf"{self}[{key}]"

        return periods



"""Export docs/examples notebooks to .py"""


def main():
    from pathlib import Path

    import nbformat
    from nbconvert import PythonExporter

    # Example: export all notebooks from docs/examples
    nb_dir = Path(__file__).parent.parent.parent / "docs" / "examples"
    print(Path(__file__).parent)
    out_dir = Path(__file__).parent / "library" / "examples"
    out_dir.mkdir(exist_ok=True)

    for nb_file in nb_dir.glob("*.ipynb"):
        with open(nb_file, "r", encoding="utf-8") as f:
            nb = nbformat.read(f, as_version=4)
        exporter = PythonExporter()
        body, _ = exporter.from_notebook_node(nb)
        out_path = out_dir / f"{nb_file.stem}.py"
        with open(out_path, "w", encoding="utf-8") as f:
            f.write(body)
        print(f"Exported {nb_file} → {out_path}")


if __name__ == "__main__":
    main()



"""dictionary utils"""

from collections import defaultdict
from operator import is_


def get_depth(d: dict) -> int:
    """
    Finds the depth of a dictionary.

    :param d: The dictionary to measure.
    :type d: dict

    :return: Depth of the dictionary.
    :rtype: int
    """
    if not isinstance(d, dict) or not d:
        # If the input is not a d or is an empty d, return 0
        return 0

    # Recursively find the maximum depth of nested dictionaries
    max_depth = max(get_depth(value) for value in d.values())

    # Return one more than the maximum depth found
    return 1 + max_depth


def flatten(d: dict, key: tuple = ()) -> dict:
    """
    Makes a flat dictionary from a nested dictionary.

    :param d: The dictionary to flatten.
    :type d: dict
    :param key: Current key path. Defaults to ``()``.
    :type key: tuple, optional

    :return: Flattened dictionary with tuple keys.
    :rtype: dict
    """
    items = []
    for key, val in d.items():
        keyupd = key + (key,)
        if isinstance(val, dict):
            items.extend(flatten(val, keyupd).items())
        else:
            items.append((keyupd, val))
    return dict(items)


def tupler(d: dict, path: tuple = ()) -> list[tuple[str]]:
    """
    Makes a list of tuples of keys in a nested dictionary.

    :param d: The dictionary to traverse.
    :type d: dict
    :param path: Current path. Defaults to ``()``.
    :type path: tuple, optional

    :return: List of tuples of keys representing the paths to each value.
    :rtype: list[tuple[str]]
    """

    result = []

    for k, v in d.items():
        path_ = path + (k,)
        result.append(path_)

        if isinstance(v, dict):
            result.extend(tupler(v, path_))

        if isinstance(v, set):
            for v_ in v:
                result.append(path_ + (v_,))

    return result


def merge_trees(d1: dict, d2: dict) -> dict:
    """
    Recursively merge two tree-like dicts (values always dicts).

    :param d1: First dictionary.
    :type d1: dict
    :param d2: Second dictionary.
    :type d2: dict

    :return: Merged dictionary.
    :rtype: dict
    """
    result = dict(d1)  # shallow copy of d1
    for k, v in d2.items():
        if k in result:
            result[k] = merge_trees(
                result[k],
                v,
            )  # recurse since v must also be a dict
        else:
            result[k] = v
    return result


def dictify(d: defaultdict | dict) -> dict:
    """
    Recursively convert defaultdict to dict.

    :param d: The dictionary to convert.
    :type d: defaultdict | dict
    """
    if isinstance(d, defaultdict):
        return {k: dictify(v) for k, v in d.items()}
    return d


class NotFoundError(Exception):
    """Raised when two periods are not linked in the period tree."""


def compare(tree: dict, of):
    """How many periods make this period"""
    n = 1
    for i, j in tree.items():
        if is_(i, of):
            return n
        n *= i.size
        if not j:
            raise NotFoundError
        n *= compare(j, of)
    return n


def dict_signature(d):
    """Hashable signature of a dict for grouping."""
    if not isinstance(d, dict):
        return None
    return tuple(
        (k, dict_signature(v)) for k, v in sorted(d.items(), key=lambda x: str(x[0]))
    )


def merge_tree_levels(d):
    """
    Convert nested dict to level-wise compact structure,
    merging repeated keys at all levels.
    """
    if not isinstance(d, dict) or not d:
        return []

    # Group children by their structure signature
    sig_to_keys = defaultdict(list)
    for k, v in d.items():
        sig_to_keys[dict_signature(v)].append((k, v))

    current_level = []
    next_levels = []

    for _, kv_list in sig_to_keys.items():
        keys = [k for k, _ in kv_list]
        children = [v for _, v in kv_list]

        # flatten singletons
        current_level.append(keys if len(keys) > 1 else keys[0])

        # recurse only once per identical children
        child_levels = merge_tree_levels(children[0])
        if child_levels:
            next_levels.append(child_levels)

    # merge next_levels by depth
    merged_next_levels = []
    for group in next_levels:
        for i, lvl in enumerate(group):
            if len(merged_next_levels) <= i:
                merged_next_levels.append([])
            # flatten duplicates at same level
            for item in lvl:
                if item not in merged_next_levels[i]:
                    merged_next_levels[i].append(item)

    return [current_level] + merged_next_levels



"""Transport"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .operation import Operation

if TYPE_CHECKING:
    # from ..commodities.resource import Resource
    from ..spatial.linkage import Linkage
    from ..spatial.location import Location

    # from ..temporal.periods import Periods


class Transport(Operation):
    """
    Exports Resource through Link basically, moves Resources between Locations

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param citations: An optional citation or description for the component. Defaults to None.
    :type citations: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversion: Operational conversion associated with the operation. Defaults to None.
    :vartype conversion: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar fab: Material conversion associated with the operation. Defaults to None.
    :vartype fab: Conversion, optional
    :ivar _fab_balanced: True if the material conversion has been balanced. Defaults to False.
    :vartype _fab_balanced: bool
    :ivar linkages: List of Linkages where the transport is located. Defaults to [].
    :vartype linkages: list[Linkage]
    """

    def __init__(self, *args, label: str = "", citations: str = "", **kwargs):

        Operation.__init__(self, *args, label=label, citations=citations, **kwargs)
        self.linkages: list[Linkage] = []

    @property
    def spaces(self) -> list[Location]:
        """Locations at which the process is balanced"""
        return self.linkages

    # def write_production(self, link_times: list[tuple[Linkage, Periods]]):
    #     """Write the conversion constraints for the transport"""

    #     def time_checker(res: Resource, loc: Location, time: Periods):
    #         """This checks if it is actually necessary
    #         to write conversion at denser temporal scales
    #         """
    #         # This checks whether some other aspect is defined at
    #         # a lower temporal scale

    #         if loc not in self.model.balances[res]:
    #             # if not defined for that location, check for a lower order location
    #             # i.e. location at a lower hierarchy,
    #             # e.g. say if loc being passed is a city, and a grb has not been defined for it
    #             # then we need to check at a higher order
    #             parent = self.space.split(loc)[1]  # get location at one hierarchy above
    #             if parent:
    #                 # if that indeed exists, then make the parent the loc
    #                 # the conversion Balance variables will feature in grb for parent location
    #                 loc = parent

    #         _ = self.model.balances[res][loc][time]

    #         #     self.model.update_balances(resource=res, space=loc, time=time)

    #         # if time not in self.model.balances[res][loc]:
    #         #     self.model.update_balances(resource=res, space=loc, time=time)
    #         if res.inv_of:
    #             # for inventoried resources, the conversion is written
    #             # using the time of the base resource's grb
    #             res = res.inv_of

    #         times = list(
    #             [
    #                 t
    #                 for t in self.model.balances[res][loc]
    #                 if self.model.balances[res][loc][t]
    #             ],
    #         )
    #         # write the conversion balance at
    #         # densest temporal scale in that space
    #         if times:
    #             return min(times)

    #         return time.horizon

    #     self.production.balancer()

    #     if self.production.pwl:

    #         conversion = self.balance[list(self.balance)[0]]

    #     else:
    #         conversion = self.balance

    #     shipping_conversion, rest_conversion = {self.production.resource: 1}, {
    #         k: v for k, v in conversion.items() if k != self.production.resource
    #     }

    #     for link_time in link_times:
    #         link, time = link_time
    #         source, sink = link.source, link.sink

    #         # time = link_time[1]

    #         if link in self.linkages:
    #             # if the transport is already balanced for the location , Skip
    #             continue

    #         for res, par in conversion.items():
    #             # set, the conversion on the resource
    #             setattr(res, self.name, self)
    #             # now there are two cases possible
    #             # the parameter (par) is positive or negative
    #             # if positive, the resource is expend
    #             # if negative, the resource is produced
    #             # also, the par can be an number or a list of numbers

    #             # insitu resource (expended and ship_outed within the system)
    #             # do not initiate a grb so we need to run a check for that first
    #             if res in self.model.balances:
    #                 time = time_checker(res, link.source, time)

    #                 if self.model.balances[res][link][time]:
    #                     # if the grb has been defined for that resource at that location and time
    #                     _insitu = False
    #                 else:
    #                     _insitu = True
    #             else:
    #                 # this implies that the grb needs to be initiated
    #                 # by declaring relevant variable
    #                 # the relevant variable will be unbounded
    #                 _insitu = True

    #             if isinstance(par, (int | float)) and par < 0:
    #                 # condition: negative number
    #                 eff = -par

    #                 if self.lag:
    #                     opr = self.operate(link, self.lag.of)
    #                     rhs = res.expend(self.operate, link.source, self.lag.of)
    #                 else:
    #                     opr = self.operate(link, time)
    #                     rhs = res.expend(opr, link.source, time)

    #             elif isinstance(par, list) and par[0] < 0:
    #                 # condition: list with negative numbers
    #                 eff = [-i for i in par]

    #                 if self.lag:
    #                     opr = self.operate(link, self.lag.of)
    #                     rhs = res.expend(self.operate, link.source, self.lag.of)
    #                 else:
    #                     opr = self.operate(link, time)
    #                     rhs = res.expend(self.operate, link.source, time)

    #             else:
    #                 # condition: positive number or list of positive numbers
    #                 eff = par

    #                 if self.lag:
    #                     opr = self.operate(link, self.lag)
    #                     rhs_export = self.model.ship_out(
    #                         res,
    #                         self.operate,
    #                         link.source,
    #                         self.lag.of,
    #                     )
    #                     rhs_import = self.model.ship_in(
    #                         res,
    #                         self.operate,
    #                         link.sink,
    #                         self.lag,
    #                     )
    #                 else:
    #                     opr = self.operate(link, time)
    #                     rhs_export = self.model.ship_out(
    #                         res,
    #                         self,
    #                         link.source,
    #                         time_checker(res, link.source, time),
    #                     )
    #                     rhs_import = self.model.ship_in(
    #                         res,
    #                         self,
    #                         link.sink,
    #                         time_checker(res, link.sink, time),
    #                     )

    #             if _insitu:
    #                 res.insitu = True
    #                 _ = rhs_export == True
    #                 _ = rhs_import == True

    #             if self.production.pwl:

    #                 eff = [conv[res] for conv in self.balance.values()]

    #                 if eff[0] < 0:
    #                     eff = [-i for i in eff]

    #                 if not self.production.modes_set:
    #                     self.model.operate.bound = None
    #                     _ = opr == dict(enumerate(self.balance.keys()))

    #                     self.model.operate.bound = self.production.model.capacity

    #                     modes = self.model.modes[-1]
    #                     self.production.modes_set = True

    #                 else:
    #                     modes = self.production.modes
    #                     modes.bind = self.operate
    #                     self.production.modes_set = True

    #                 opr = opr(modes)
    #                 rhs_export = rhs_export(modes)
    #                 rhs_import = rhs_import(modes)

    #             _ = opr[rhs_export] == eff
    #             _ = opr[rhs_import] == eff

    #             self.linkages.append(link)



"""A Pre-set Parameter"""

from __future__ import annotations

from typing import TYPE_CHECKING, Type

from ..._core._name import _Name

if TYPE_CHECKING:
    from ..._core._component import _Component
    from ...represent.model import Model


class Instruction(_Name):
    """
    Pre-set instructions to deal with parameter calculations

    :param deciding: Name of the deciding aspect
    :type deciding: str
    :param depending: Name of the depending aspect
    :type depending: str
    :param default: Name of the default component
    :type default: str
    :param label: Label for the parameter. Defaults to ''.
    :type label: str, optional
    :param latex: LaTeX representation of the parameter. Defaults to ''.
    :type latex: str, optional

    :raises TypeError: If the component provided is not of the correct kind.
    """

    def __init__(
        self,
        name: str,
        kind: Type[_Component],
        deciding: str,
        depending: str,
        default: str,
        label: str = "",
        latex: str = "",
    ):

        _Name.__init__(self, label=label)
        self.name = name
        self.kind = kind
        self.deciding = deciding
        self.depending = depending
        self.default = default
        self.model: Model | None = None
        self.latex = latex or self.name

    def __call__(self, component: _Component):
        # create the calculation
        # there has to be a model

        if isinstance(component, self.kind):

            self.model = component.model

            default_comp = self.model.default_components[self.default]()

            depending = getattr(default_comp, self.depending)
            deciding = getattr(component, self.deciding)

            return deciding[depending]

        raise TypeError(
            f"{component} ({type(component)}): {self.name} earmarked for {self.kind.__name__}"
        )



"""Interact"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .player import Player


class Interact:
    """Interaction between Players
    
    :param player1: First player
    :type player1: Player
    :param player2: Second player
    :type player2: Player
    """

    player1: Player = None
    player2: Player = None



"""Map"""

from __future__ import annotations

import logging
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING

from gana import sigma

from ...utils.decorators import timer

logger = logging.getLogger("energia")

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ..._core._x import _X
    from ..indices.domain import Domain
    from ..variables.aspect import Aspect


class Map:
    """
    Maps between domains

    :param aspect: Aspect to which the constraint is applied.
    :type aspect: Aspect
    :param domain: Domain over which the aspect is defined.
    :type domain: Domain
    :param reporting: If True, the map is for a reporting variable.
    :type reporting: bool
    :param label: Label for the constraint. Defaults to "".
    :type label: str
    """

    def __init__(
        self, aspect: Aspect, domain: Domain, reporting=False, label: str = ""
    ):

        self.aspect = aspect
        self.domain = domain
        self.reporting = reporting
        self.label = label

        self._handshake()

        if self.domain.lag:
            return

        # these are periods denser and sparser than the current domain
        self._map_across_time()

        # Space mapping ---
        self._map_across_space()
        # Bind mapping ---
        self._map_across_samples()

        # Mode mapping ---
        self._map_across_modes()

    # -------------------------------------------------------------------#
    # Helper functions
    # -------------------------------------------------------------------#
    @timer(logger, kind='map')
    def write(self, from_domain: Domain, to_domain: Domain, tsum=False, msum=False):
        """Scales up variable to a lower dimension"""
        # if the variable is being defined for the first time, do not bother with the rest
        # Also note that if a variable already exists then a new is not created
        # thus map_domain is not called in Bind.V()
        # Essentially, the domain being added is always new!

        # 1. a new domain is being added to a new disposition
        # e.g.:
        #   (water, dam, goa, year) with disposition ['resource', 'operation', 'space', 'time'] exists
        #   (dam, goa, q) with disposition ['operation', 'space', 'time'] is added

        # # 2. a new domain is being added to an existing disposition
        # # e.g.:
        # #   (dam, goa, year) with disposition ['operation', 'space', 'time'] exists
        # #   (dam, goa, q) with disposition ['operation', 'space', 'time'] is added
        # #   The variable will need to be mapped across 'time

        # mapping happens between:
        # 1. sparser time to denser time in same space
        # 2. contained location to parent location
        # 3. (sparser time, contained location) to (denser time, parent location)
        # you cant map at the same level
        # consider the State Goa, with two towns Madgaon and Ponje
        # Goa with Maharashtra, make up India (for the purpose of modeling)
        # Aspect from Goa can be mapped to India
        # Goa does not need to be mapped to Maharashtra
        # Madgaon and Ponje are both mapped to Just Goa
        # note that .alsohas() checks at lower levels
        # the operators <, > wont work since they will add any space a higher order
        # this will lead to adding twice. India = Goa + Madgaon + Ponje (WRONG)
        # We scale only one level up
        if self._check_validity():
            return

        exists = self._check_existing(to_domain, from_domain)

        self.cons_name = self._give_cname(self.var, from_domain, to_domain, tsum, msum)

        v_lower = self(*to_domain).X() if self.reporting else self(*to_domain).V()

        if not tsum and not msum and exists:
            cons_existing: C = getattr(self.program, self.cons_name)
            setattr(
                self.program,
                self.cons_name,
                cons_existing - self.rhs(from_domain, tsum, msum),
            )

        else:
            cons = v_lower == self.rhs(from_domain, tsum, msum)
            setattr(self.program, self.cons_name, cons)
            cons.categorize("Mapping")

        self._inform(from_domain)

        return (self.aspect, from_domain, to_domain)

    @cached_property
    def var(self):
        return self.aspect.reporting if self.reporting else self.aspect

    @cached_property
    def name(self) -> str:
        return f"{self.aspect.name}_map"

    @cached_property
    def maps(self) -> dict[Domain, list[Domain]]:
        return self.aspect.maps_report if self.reporting else self.aspect.maps

    def _map_across_time(self):
        """
        Maps across time


        :param sparser_periods: Periods sparser than the domain period
        :type sparser_periods: list[Periods]
        :param denser_periods: Periods denser than the domain period
        :type denser_periods: list[Periods]
        """
        denser_periods, sparser_periods = self.model.time.split(self.time)

        for sp in sparser_periods:
            # check if the aspect has been defined for a sparser period
            # this creates a map from this domain to a sparser domain
            if sp in self.dispositions[self.space] and is_(sp.of, self.time):
                self.write(self.domain, self.domain.change({"periods": sp}), tsum=True)

        for dp in denser_periods:
            if dp in self.dispositions[self.space] and is_(self.time.of, dp):
                binds_dict = self.dispositions[self.space][dp]
                # TODO - check this
                # here I am re creating Bind objects
                # from the dict of the form {aspect: {component: {aspect: {component: {...}}}}}
                # there has to be a way to avoid this
                # I make a list  of bounds as such [aspect(component), aspect(component), ...]
                samples = [
                    aspect(component)
                    for aspect, comp_dict in binds_dict.items()
                    for component in comp_dict
                ]
                from_domain = self.domain.copy()
                from_domain.periods, from_domain.samples = dp, samples
                self.write(from_domain, self.domain, tsum=True)

    def _map_across_space(self):
        """
        Maps across space

        :param contained_locs: Locations contained in the domain location
        :type contained_locs: list[Space]
        :param parent_loc: Parent location of the domain location
        :type parent_loc: Space | None
        :param time: Time at which the domain is defined
        :type time: Periods
        """
        parent_loc = self.domain.location.isin

        if parent_loc:
            if (
                parent_loc not in self.dispositions
                or self.time not in self.dispositions[parent_loc]
            ):
                return
            self.write(
                self.domain,
                self.domain.change({"location": parent_loc}),
            )

        # contained_locs, parent_loc = self.model.space.split(self.space)
        # for location in contained_locs:
        #     if location not in dispositions or time not in dispositions[location]:
        #         # the aspect is already defined at this location
        #         # we need to check if it is defined for any periods sparser than time
        #         continue
        #     # samples = dispositions[location][time]
        #     # if not samples:

        #     self.writecons_map(self.domain.change({"location": location}), self.domain)
        # else:
        #     new_binds = [k(list(v)[0]) for k, v in samples.items()]
        #     # consider the case where overall consumption for water in some location and time is defined
        #     # now user defines consumption due to using cement during construction
        #     # we should have the constraint consume(water, goa, 2025) = consume(water, goa, 2025, use, cement)
        #     # in that case, samples is just []
        #     self.writecons_map(
        #         self.domain.change({"location": location, "samples": new_binds}),
        #         self.domain,
        #     )

        # if parent_loc in dispositions:

        #     self.writecons_map(
        #         self.domain, self.domain.change({"location": parent_loc})
        #     )

    def _map_across_samples(self):
        if self.domain.samples or not self.dispositions[self.space][self.time]:
            return
        # if the current variable being declared has no samples
        # but the aspect has already been defined at this location and time with samples
        # there is a need to map from the defined samples to no samples
        # get list of domains of the aspect

        domains = [
            d
            for d in self.aspect.domains
            if is_(d.primary, self.domain.primary)
            and is_(d.space, self.space)
            and is_(d.periods, self.time)
            and d.samples
        ]
        for domain in domains:
            if domain.modes is None:
                self.write(domain, self.domain)

    def _map_across_modes(self):
        # modes need some additional checks
        # constraints could be written using parent modes or child modes
        # these checks avoid writing the same constraint twice
        # AVOIDS: v_t = v_t,m1 + v_t,m2 + v_t,m; v_t,m = v_t,m1 + v_t,m2
        # CORRECT: v_t = v_t,m; v_t,m = v_t,m1 + v_t,m2 OR v_t = v_t,m1 + v_t,m2
        if not self.domain.modes:
            return

        if not self._special_modes_check(
            self.domain, self.domain.change({"modes": None})
        ):
            return

        if self.domain.modes.parent:
            self.write(self.domain, self.domain.change({"modes": None}))
        else:
            self.write(self.domain, self.domain.change({"modes": None}), msum=True)

    def _check_existing(self, to_domain: Domain, from_domain: Domain) -> bool:
        """Checks if the map constraint already exists"""
        what = (to_domain - from_domain)[0]
        if to_domain not in self.maps[what]:
            self.maps[what][to_domain] = [from_domain]
            # make new constraint
            return False

        if from_domain in self.maps[what][to_domain]:
            # There is an existing map
            return True
        else:
            self.maps[what][to_domain].append(from_domain)
        if what in ["samples", "modes"]:
            return True

    def _special_modes_check(self, from_domain: Domain, to_domain: Domain) -> None:
        """
        Checks if constraints have been written for child modes
        or for parent modes
        """
        if to_domain in self.aspect.maps["modes"]:
            if from_domain.modes.parent:
                if (
                    from_domain.change({"modes": from_domain.modes.parent})
                    in self.aspect.maps["modes"][to_domain]
                ):
                    return False
            else:
                for modes in from_domain.modes._:
                    if (
                        from_domain.change({"modes": modes})
                        in self.aspect.maps["modes"][to_domain]
                    ):
                        return False
        return True

    def _give_cname(
        self,
        var,
        from_domain,
        to_domain,
        tsum: bool = False,
        msum: bool = False,
    ) -> str:
        """Return canonical map constraint name based on domain relationship and aggregation type."""

        if tsum:
            return f"{var}{from_domain.idxname}_to_{to_domain.idxname}_tmap"

        if msum:
            # Original behavior: use from_domain idxname for per-mode naming
            return f"{var}{from_domain.idxname}_mmap"

        if from_domain.modes:
            if from_domain.modes.parent:
                parent_domain = from_domain.change({"modes": from_domain.modes.parent})
                return f"{var}{parent_domain.idxname}_mmap"
            return f"{var}{from_domain.idxname}_mmap"

        return f"{var}{to_domain.idxname}_map"

    def rhs(self, domain: Domain, tsum=False, msum=False):
        """Gives the sum of the variable over the domain"""
        varname = (
            f"x_{self.aspect.name}" if (msum and self.reporting) else self.aspect.name
        )
        v = getattr(self.program, varname)

        if tsum:
            # if the domain has been mapped to but this is a time sum
            # we need to first map time
            # and then add it to an existing map at a lower domain
            return sigma(v(*domain.I), domain.time.i)
        if msum:
            # if the domain has been mapped to but this is a mode sum
            # we need to first map modes
            # and then add it to an existing map at a lower domain
            return sigma(v(*domain.I), domain.modes.I)
        # the copy is important since otherwise, the printing will take
        # the update index if the variable is mutated

        return v(*domain.I).copy()

    # -------------------------------------------------------------------#
    # Constraint writing
    # -------------------------------------------------------------------#

    def _check_validity(self) -> bool:
        """Check space and time"""
        if (
            self.space not in self.dispositions
            or self.time not in self.dispositions[self.space]
        ):
            return True

    def _inform(self, from_domain: Domain):
        if self.cons_name not in self.aspect.constraints:
            self.aspect.constraints.append(self.cons_name)

        from_domain.inform_indices(self.cons_name)

    def _handshake(self):
        """Borrow attributes from aspect"""
        self.model = self.aspect.model
        # these are spaces contained in location and parent location to which this location belongs
        # this gives all the dispositions at which the aspect has been defined
        self.dispositions = self.model.dispositions[self.aspect][self.domain.primary]
        self.program = self.model.program

        # this is the disposition of the variable to be mapped
        # through time and space
        self.time, self.space = self.domain.periods, self.domain.space

    def __call__(self, *index: _X):
        return self.aspect(*index)



"""Control Variable"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Self

from ..constraints.vmap import Map
from .aspect import Aspect

if TYPE_CHECKING:
    from ..indices.domain import Domain
    from .states import State


@dataclass
class Control(Aspect):
    """
    Some action to be performed by a component
    that adjusts a state variable
    can have a consequence (Impact)
    or elicit a motion (Stream) that has a consequence (Impact)

    :param primary_types: associated components type(s). Defaults to None.
    :type primary_types: Type[_Component] | tuple[Type[_Component], ...]
    :param nn: If True, the decision is a non-negative decision. Defaults to True.
    :type nn: bool
    :param ispos: If True, the decision is positive (non-negative). Defaults to True.
    :type ispos: bool
    :param neg: Negative form or representation of the decision, if any. Defaults to "".
    :type neg: str
    :param latex: LaTeX string. Defaults to "".
    :type latex: str
    :param bound: if the aspect is bounded by another. Defaults to "".
    :type bound: str
    :param label: Label for the decision. Defaults to "".
    :type label: str

    :ivar model: Model to which the Aspect belongs.
    :vartype model: Model
    :ivar name: Name of the Aspect.
    :vartype name: str
    :ivar indices: List of indices (Location, Periods) associated with the Aspect.
    :vartype indices: list[Location | Linkage, Periods]
    :ivar bound_spaces: Spaces where the Aspect has been already bound.
    :vartype bound_spaces: dict[Commodity | Process | Storage | Transport, list[Location | Linkage]]
    :ivar domains: List of domains associated with the Aspect.
    :vartype domains: list[Domain]

    :raises ValueError: If `primary_type` is not defined.
    """

    def __post_init__(self):
        Aspect.__post_init__(self)

        # a state that is being controlled
        self.state: State = None
        # a stream that is being controlled
        self.stream: Self = None
        # an impact that is being controlled
        self.impact: Self = None
        # what increase the state/impact/stream
        self.more: Self = None
        # what decreases the state/impact/stream
        self.less: Self = None

        self.mapped_from: list[Domain] = []

    def update(self, domain: Domain, reporting: bool = False):
        """Add a domain to the decision variable"""

        # Write a mapping constraint
        self.Map(domain=domain, reporting=reporting)



"""Energy system examples"""

from ...components.commodities.currency import Currency
from ...components.commodities.material import Material
from ...components.commodities.resource import Resource
from ...components.impact.categories import Environ
from ...components.measure.unit import Unit
from ...components.operations.process import Process
from ...components.operations.storage import Storage
from ...components.spatial.location import Location
from ...components.temporal.periods import Periods
from ...components.temporal.scales import TemporalScales
from ...represent.model import Model


def scheduling():
    """A small scheduling example"""
    m = Model("scheduling")
    m.q = Periods()
    m.y = 4 * m.q
    m.usd = Currency()
    m.power, m.wind = Resource(), Resource()
    _ = m.wind.consume <= 400
    _ = m.power.release.prep(100) >= [0.6, 0.7, 1, 0.3]
    m.wf = Process()
    _ = m.wf(m.power) == -1 * m.wind
    _ = m.wf.operate.prep(200, norm=False) <= [0.9, 0.8, 0.5, 0.7]
    _ = m.usd.spend(m.wf.operate) == [4000, 4200, 4300, 3900]
    m.network.locate(m.wf)

    return m


def design_scheduling_w_gattr():
    """A small design and scheduling example"""
    m = Model("design_scheduling")
    m.q = Periods()
    m.y = 4 * m.q
    m.usd = Currency()
    m.declare(Resource, ["power", "wind", "solar"])
    _ = m.solar.consume(m.q) <= 100
    _ = m.wind.consume <= 400
    _ = m.power.release.prep(180) >= [0.6, 0.7, 0.8, 0.3]
    m.wf = Process()
    _ = m.wf(m.power) == -1 * m.wind
    _ = m.wf.capacity.x <= 100
    _ = m.wf.capacity.x >= 10
    _ = m.wf.operate.prep(norm=True) <= [0.9, 0.8, 0.5, 0.7]
    _ = m.wf.capacity[m.usd.spend] == 990637 + 3354
    _ = m.wf.operate[m.usd.spend] == 49

    m.pv = Process()
    _ = m.pv(m.power) == -1 * m.solar
    _ = m.pv.capacity.x <= 100
    _ = m.pv.capacity.x >= 10
    _ = m.pv.operate.prep(norm=True) <= [0.6, 0.8, 0.9, 0.7]
    _ = m.pv.capacity[m.usd.spend] == 567000 + 872046
    _ = m.pv.operate[m.usd.spend] == 90000

    m.lii = Storage()
    _ = m.lii(m.power) == 0.9
    _ = m.lii.capacity.x <= 100
    _ = m.lii.capacity.x >= 10
    _ = m.lii.capacity[m.usd.spend] == 1302182 + 41432
    _ = m.lii.inventory[m.usd.spend] == 2000

    m.network.locate(m.wf, m.pv, m.lii)

    return m


def design_scheduling():
    """
    A small design and scheduling example
    with functions for calculations instead of getattr
    """
    m = Model("design_scheduling")
    m.q = Periods()
    m.y = 4 * m.q
    m.usd = Currency()
    m.declare(Resource, ["power", "wind", "solar"])
    _ = m.solar.consume(m.q) <= 100
    _ = m.wind.consume <= 400
    _ = m.power.release.prep(180) >= [0.6, 0.7, 0.8, 0.3]

    m.wf = Process(m.power == -1 * m.wind)
    _ = m.wf.capacity.x <= 100
    _ = m.wf.capacity.x >= 10
    _ = m.wf.operate.prep(norm=True) <= [0.9, 0.8, 0.5, 0.7]
    _ = m.usd.spend(m.wf.capacity) == 990637 + 3354
    _ = m.usd.spend(m.wf.operate) == 49

    m.pv = Process(m.power == -1 * m.solar)
    _ = m.pv.capacity.x <= 100
    _ = m.pv.capacity.x >= 10
    _ = m.pv.operate.prep(norm=True) <= [0.6, 0.8, 0.9, 0.7]
    _ = m.usd.spend(m.pv.capacity) == 567000 + 872046
    _ = m.usd.spend(m.pv.operate) == 90000

    m.lii = Storage(m.power == 0.9)
    _ = m.lii.capacity.x <= 100
    _ = m.lii.capacity.x >= 10
    _ = m.usd.spend(m.lii.capacity) == 1302182 + 41432
    _ = m.usd.spend(m.lii.inventory) == 2000
    m.network.locate(m.wf, m.pv, m.lii)

    return m


def design_scheduling_w_attrs():
    """A small design and scheduling example with attributes"""

    m = Model()
    m.q = Periods()
    m.y = 4 * m.q
    m.usd = Currency()
    m.solar = Resource(consume_max=[100] * 4)
    m.wind = Resource(consume_max=100 * 4)
    m.power = Resource(demand_nominal=180, demand_min=[0.6, 0.7, 0.8, 0.3])
    m.wf = Process(
        m.power == -1.0 * m.wind,
        capacity_max=100,
        capacity_min=10,
        capacity_optional=True,
        operate_normalize=True,
        operate_max=[0.9, 0.8, 0.5, 0.7],
        capex=990637 + 3354,
        opex=49,
    )

    m.pv = Process(
        m.power == -1 * m.solar,
        capacity_max=100,
        capacity_min=10,
        capacity_optional=True,
        operate_normalize=True,
        operate_max=[0.6, 0.8, 0.9, 0.7],
        capex=567000 + 872046,
        opex=90000,
    )
    m.lii = Storage(
        m.power == 0.9,
        capacity_max=100,
        capacity_min=10,
        capacity_optional=True,
        capex=1302182 + 41432,
        inventory_cost=2000,
    )

    m.locate(m.wf, m.pv, m.lii)

    return m


def design_scheduling_material():

    m = Model("design_scheduling_material")
    m.q = Periods()
    m.y = 4 * m.q
    m.usd = Currency()

    m.declare(Resource, ['power', 'wind', 'solar'])
    _ = m.solar.consume(m.q) <= 100
    _ = m.wind.consume <= 400
    _ = m.power.release.prep(180) >= [0.6, 0.7, 0.8, 0.3]

    m.wf = Process()
    _ = m.wf(m.power) == -1 * m.wind
    _ = m.wf.capacity.x <= 100
    _ = m.wf.capacity.x >= 10
    _ = m.wf.operate.prep(norm=True) <= [0.9, 0.8, 0.5, 0.7]
    _ = m.usd.spend(m.wf.capacity) == 990637 + 3354
    _ = m.usd.spend(m.wf.operate) == 49

    m.pv = Process()
    _ = m.pv(m.power) == -1 * m.solar
    _ = m.pv.capacity.x <= 100
    _ = m.pv.capacity.x >= 10
    _ = m.pv.operate.prep(norm=True) <= [0.6, 0.8, 0.9, 0.7]
    _ = m.usd.spend(m.pv.capacity) == 567000 + 872046
    _ = m.usd.spend(m.pv.operate) == 90000

    m.lii = Storage()
    _ = m.lii(m.power) == 0.9
    _ = m.lii.capacity.x <= 100
    _ = m.lii.capacity.x >= 10
    _ = m.usd.spend(m.lii.capacity) == 1302182 + 41432
    _ = m.usd.spend(m.lii.inventory) == 2000

    m.cement = Material()
    _ = m.cement.consume <= 1000000
    _ = m.usd.spend(m.cement.consume) == 17
    _ = m.cement.use(m.wf.capacity) == 400
    _ = m.cement.use(m.pv.capacity) == 560
    _ = m.cement.use(m.lii.capacity) == 300

    m.gwp = Environ()
    _ = m.gwp.emit(m.wf.capacity) == 1000
    _ = m.gwp.emit(m.pv.capacity) == 2000
    _ = m.gwp.emit(m.lii.capacity) == 3000
    _ = m.gwp.emit(m.cement.consume) == 0.9

    m.network.locate(m.wf, m.pv, m.lii)

    return m


def design_scheduling_material_modes():
    """Design and scheduling considering materials and emissions from them"""
    m = Model("design_scheduling")
    m.scales = TemporalScales([1, 4], ["y", "q"])
    m.usd = Currency()
    m.gwp = Environ()
    m.declare(
        Material,
        [
            "lir",
            "lib",
            "steel",
            "concrete",
            "glass",
            "si_mono",
            "si_poly",
        ],
    )

    m.declare(Resource, ["power", "wind", "solar"])
    _ = m.solar.consume == True
    _ = m.wind.consume == True
    _ = m.power.release.prep(180) >= [0.6, 0.7, 0.8, 0.3]

    _ = m.gwp.emit(m.lir.consume) == 9600
    _ = m.gwp.emit(m.lib.consume) == 2800
    _ = m.gwp.emit(m.steel.consume) == 2121.152427
    _ = m.usd.spend(m.steel.consume) == 670
    _ = m.gwp.emit(m.concrete.consume) == 120.0378
    _ = m.gwp.emit(m.glass.consume) == 1118.5
    _ = m.gwp.emit(m.si_mono.consume) == 122239.1
    _ = m.gwp.emit(m.si_poly.consume) == 98646.7

    # _ = m.lir.consume[m.gwp.emit] == 9600
    # _ = m.lib.consume[m.gwp.emit] == 2800
    # _ = m.steel.consume[m.gwp.emit] == 2121.152427
    # _ = m.steel.consume[m.usd.spend] == 670
    # _ = m.concrete.consume[m.gwp.emit] == 120.0378
    # _ = m.glass.consume[m.gwp.emit] == 1118.5
    # _ = m.si_mono.consume[m.gwp.emit] == 122239.1
    # _ = m.si_poly.consume[m.gwp.emit] == 98646.7

    m.wf = Process()
    _ = m.wf.capacity.x <= 100
    _ = m.wf.capacity.x >= 10
    _ = m.wf.operate.prep(norm=True) <= [0.9, 0.8, 0.5, 0.7]
    _ = m.wf.construction == [
        -109.9 * m.steel - 398.7 * m.concrete,
        -249.605 * m.steel - 12.4 * m.concrete,
    ]

    _ = m.wf(m.power) == {
        m.wf.construction.modes[0]: -2.857 * m.wind,
        m.wf.construction.modes[1]: -2.3255 * m.wind,
    }

    _ = m.usd.spend(m.wf.capacity, m.wf.construction.modes) == [
        1292000 + 29200,
        3192734 + 101498,
    ]

    _ = m.usd.spend(m.wf.operate) == 49

    m.pv = Process()
    _ = m.pv.construction == [
        -70 * m.glass - 7 * m.si_mono,
        -70 * m.glass - 7 * m.si_poly,
    ]

    _ = m.pv(m.power) == {
        m.pv.construction.modes[0]: -5 * m.solar,
        m.pv.construction.modes[1]: -6.67 * m.solar,
    }
    _ = m.pv.capacity.x <= 100
    _ = m.pv.capacity.x >= 10
    _ = m.pv.operate.prep(norm=True) <= [0.6, 0.8, 0.9, 0.7]

    _ = m.usd.spend(m.pv.capacity) == 567000 + 872046
    _ = m.usd.spend(m.pv.operate) == 90000

    m.lii = Storage()
    _ = m.lii(m.power) == 0.9
    _ = m.lii.construction == [
        -0.137 * m.lib - 1.165 * m.steel,
        -0.137 * m.lir - 1.165 * m.steel,
    ]
    _ = m.lii.capacity.x <= 100
    _ = m.lii.capacity.x >= 10

    _ = m.usd.spend(m.lii.capacity) == 1302182 + 41432
    _ = m.usd.spend(m.lii.inventory) == 2000

    m.network.locate(m.wf, m.pv, m.lii)
    return m


def supermarket():
    """A supermarket energy system example

    Optimize for cost:
        m.usd.spend.opt()

    Optimize for CO2 emissions:
        m.co2_vent.release.opt()

    Maximize efficiency:
        m._.lb(sum(m._.consume))
    """
    resource_demand_dict = {
        "Lighting": 200,
        "Refrigeration": 1000,
        "Space Heating": 100,
    }

    resource_dict = {
        "Price": {"Natural Gas": 8.89, "Biomass": 9.72, "Grid Electricity": 36.11},
        "CO2 Generation": {"Natural Gas": 56, "Biomass": 100, "Grid Electricity": 90},
    }

    generation_process_dict = {
        "Biomass ST": {
            "nE": 68,
            "nH": 0,
            "LB": 100,
            "UB": 10**6,
            "Capex": 250,
            "Opex": 15,
        },
        "Natural Gas CHP": {
            "nE": 44,
            "nH": 28,
            "LB": 800,
            "UB": 10**6,
            "Capex": 500,
            "Opex": 15,
        },
        "Solar PV": {"nE": 9, "nH": 0, "LB": 10, "UB": 300, "Capex": 2000, "Opex": 500},
        "Wind Farm": {
            "nE": 22,
            "nH": 0,
            "LB": 10,
            "UB": 500,
            "Capex": 2000,
            "Opex": 1200,
        },
    }

    consumption_process_dict = {
        "Refrigeration": {"Efficiency": 300, "Capex": 80, "Opex": 85},
        "LED": {"Efficiency": 70, "Capex": 10, "Opex": 30},
        "Heating": {"Efficiency": 4, "Capex": 1, "Opex": 3},
    }

    m = Model("supermarket")

    m.supermarket = Location()

    m.GJ = Unit(label="Giga Joules")
    m.PJ = 10**6 * m.GJ
    m.kW = Unit(label="kilo Watts")
    m.tons = Unit(label="US Tons")

    # Resources can be declared along with thier attributes such as maximum allowed consumption, dischargeablity, base price, etc.

    m.usd = Currency(label="$")

    m.biomass = Resource(basis=m.GJ, label="Biomass")
    _ = m.biomass.consume == True
    _ = m.biomass.consume[m.usd.spend] == resource_dict["Price"]["Biomass"]

    m.gridpower = Resource(basis=m.GJ, label="Grid Electricity")
    _ = m.gridpower.consume == True
    _ = m.gridpower.consume[m.usd.spend] == resource_dict["Price"]["Grid Electricity"]

    m.ng = Resource(basis=m.GJ, label="Natural gas")
    _ = m.ng.consume == True
    _ = m.ng.consume[m.usd.spend] == resource_dict["Price"]["Natural Gas"]

    m.power = Resource(basis=m.kW, label="Electrical Power")
    m.heat = Resource(basis=m.kW, label="Heat Power")

    _ = m.co2_vent = Resource(basis=m.tons, label="Carbon dioxide")
    _ = m.co2_vent.release == True

    m.solar = Resource(basis=m.GJ, label="Solar energy")
    _ = m.solar.consume == True
    _ = m.wind = Resource(basis=m.GJ, label="Wind energy")
    _ = m.wind.consume == True

    m.lighting = Resource(basis=m.kW, label="Lighting")
    _ = m.lighting.release >= resource_demand_dict["Lighting"]

    m.refrigeration = Resource(basis=m.kW, label="Refrigeration")
    _ = m.refrigeration.release >= resource_demand_dict["Refrigeration"]

    m.heating = Resource(basis=m.kW, label="Heating")
    _ = m.heating.release >= resource_demand_dict["Space Heating"]

    m.st = Process(label="Biomass ST")
    _ = (
        m.st(-m.biomass)
        == (277.78 * generation_process_dict["Biomass ST"]["nE"] / 100) * m.power
        + m.co2_vent * resource_dict["CO2 Generation"]["Biomass"]
    )
    # using x sets a binary, making the setting up of the process optional
    _ = m.st.capacity.x >= generation_process_dict["Biomass ST"]["LB"]
    _ = m.st.capacity.x <= generation_process_dict["Biomass ST"]["UB"]
    _ = (
        m.st.capacity[m.usd.spend]
        == generation_process_dict["Biomass ST"]["Capex"] * 0.05
    )
    _ = m.st.operate[m.usd.spend] == generation_process_dict["Biomass ST"]["Opex"]

    m.chp = Process(label="Biomass ST")
    _ = (
        m.chp(-m.ng)
        == (277.78 * generation_process_dict["Natural Gas CHP"]["nE"] / 100) * m.power
        + (277.78 * generation_process_dict["Natural Gas CHP"]["nH"] / 100) * m.heat
        + m.co2_vent * resource_dict["CO2 Generation"]["Natural Gas"]
    )
    # using x sets a binary, making the setting up of the process optional
    _ = m.chp.capacity.x >= generation_process_dict["Natural Gas CHP"]["LB"]
    _ = m.chp.capacity.x <= generation_process_dict["Natural Gas CHP"]["UB"]
    _ = (
        m.chp.capacity[m.usd.spend]
        == generation_process_dict["Natural Gas CHP"]["Capex"] * 0.05
    )
    _ = m.chp.operate[m.usd.spend] == generation_process_dict["Natural Gas CHP"]["Opex"]

    m.pv = Process(label="Solar PV")
    _ = (
        m.pv(-m.solar)
        == (277.78 * generation_process_dict["Solar PV"]["nE"] / 100) * m.power
    )
    _ = m.pv.capacity.x <= generation_process_dict["Solar PV"]["UB"]
    _ = m.pv.capacity.x >= generation_process_dict["Solar PV"]["LB"]
    _ = (
        m.pv.capacity[m.usd.spend]
        == generation_process_dict["Solar PV"]["Capex"] * 0.05
    )
    _ = m.pv.operate[m.usd.spend] == generation_process_dict["Solar PV"]["Opex"]

    m.wf = Process(label="Wind Farm")
    _ = (
        m.wf(-m.wind)
        == (277.78 * generation_process_dict["Wind Farm"]["nE"] / 100) * m.power
    )
    _ = m.wf.capacity.x <= generation_process_dict["Wind Farm"]["UB"]
    _ = m.wf.capacity.x >= generation_process_dict["Wind Farm"]["LB"]
    _ = (
        m.wf.capacity[m.usd.spend]
        == generation_process_dict["Wind Farm"]["Capex"] * 0.05
    )
    _ = m.wf.operate[m.usd.spend] == generation_process_dict["Wind Farm"]["Opex"]

    m.grid = Process(label="Grid Electricity")
    _ = (
        m.grid(-m.gridpower)
        == 277.78 * m.power
        + m.co2_vent * resource_dict["CO2 Generation"]["Grid Electricity"]
    )
    _ = m.grid.capacity <= 10**5  # no binary needed because no upper bound

    m.refrigerator = Process(label="Refrigerator")
    _ = m.refrigerator.capacity <= 10**5
    _ = (
        m.refrigerator(-m.power)
        == (consumption_process_dict["Refrigeration"]["Efficiency"] / 100)
        * m.refrigeration
    )
    _ = (
        m.refrigerator.capacity[m.usd.spend]
        == consumption_process_dict["Refrigeration"]["Capex"] * 0.05
    )
    _ = (
        m.refrigerator.operate[m.usd.spend]
        == consumption_process_dict["Refrigeration"]["Opex"]
    )

    m.led = Process(label="LED")
    _ = (
        m.led(-m.power)
        == (consumption_process_dict["LED"]["Efficiency"] / 100) * m.lighting
    )
    _ = m.led.capacity[m.usd.spend] == consumption_process_dict["LED"]["Capex"] * 0.05
    _ = m.led.operate[m.usd.spend] == consumption_process_dict["LED"]["Opex"]

    m.heater = Process(label="Heater")
    _ = (
        m.heater(-m.heat)
        == (consumption_process_dict["Heating"]["Efficiency"] / 100) * m.heating
    )
    _ = (
        m.heater.capacity[m.usd.spend]
        == consumption_process_dict["Heating"]["Capex"] * 0.05
    )
    _ = m.heater.operate[m.usd.spend] == consumption_process_dict["Heating"]["Opex"]

    # Locations serve as aggregations of scenarios

    m.supermarket.locate(
        m.st,
        m.chp,
        m.pv,
        m.wf,
        m.grid,
        m.refrigerator,
        m.led,
        m.heater,
    )

    # Various constraints are need for the formulating the mathematical programming model. Here we include the cost, production, resource balance, demand, inventory, and network constraints.

    return m



"""Graph"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._hash import _Hash
from ...components.graph.edge import Edge
from ...components.graph.node import Node

if TYPE_CHECKING:
    from ..model import Model


@dataclass
class Graph(_Hash):
    """Graph representation

    :param model: Model to which the graph belongs.
    :type model: Model

    :ivar name: Name of the graph. Defaults to None.
    :vartype name: str
    :ivar nodes: List of nodes in the graph.
    :vartype nodes: list[Node]
    :ivar edges: List of edges in the graph.
    :vartype edges: list[Edge]

    .. note::
        - name is generated based on Model name
        - nodes and edges are populated as model is defined
    """

    model: Model

    def __post_init__(self):
        self.name = f"Graph({self.model})"
        self.nodes: list[Node] = []
        self.edges: list[Edge] = []
